#!/usr/bin/env python

# Standard library imports
import os
import sys

from contextlib import redirect_stdout

try:
    from importlib.resources import files, as_file
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    from importlib_resources import files, as_file

import numpy as np
import pandas as pd
import scipy.stats as stats

from sklearn_genetic.space import Continuous, Categorical, Integer

from pgsui import GenotypeData
from pgsui.impute.estimators import *
from pgsui.impute.simple_imputers import *
from pgsui.example_data import structure_files, phylip_files, popmaps, trees


def main():
    """Test all PG-SUI Methods.

    Can be invoked by typing 'pgsuitest' on the command line.
    """

    # Redirect stdout to logfile
    with open("pgsuitest.log.txt", "w") as logfile:
        with redirect_stdout(logfile):
            testaln = {
                "phylip": "test_n10.phy",
                "structure2row": "test.nopops.2row.10sites.str",
                "structure2rowPopID": "test.pops.2row.10sites.str",
                "structure1row": "test.nopops.1row.10sites.str",
                "structure1rowPopID": "test.pops.1row.10sites.str",
            }

            popmap = "test.popmap"
            tre = "test.tre"
            iqtre = "test.iqtree"
            qmat = "test.qmat"
            siterateiqtree = "test_n10.rate"
            siterate = "test_siterates_n10.txt"
            prefix = "setuptest"
            t = ".xxinput_treexx.tre"  # temporary treefile

            strfile = files(structure_files).joinpath(testaln["structure2row"])
            popmapfile = files(popmaps).joinpath(popmap)
            treefile = files(trees).joinpath(tre)
            iqtreeqmatfile = files(trees).joinpath(iqtre)
            qmatfile = files(trees).joinpath(qmat)
            siteratefileiqtree = files(trees).joinpath(siterateiqtree)
            siteratefile = files(trees).joinpath(siterate)

            with as_file(popmapfile) as m, as_file(
                treefile
            ) as guidetree, as_file(iqtreeqmatfile) as i, as_file(
                qmatfile
            ) as q, as_file(
                siteratefileiqtree
            ) as siq, as_file(
                siteratefile
            ) as s:

                # Added this code block because for some reason toytree won't
                # read the as_file() temporary file using the context manager.
                with open(guidetree, "r") as fin:
                    input_tree = fin.read()
                    with open(t, "w") as fout:
                        fout.write(input_tree)

                print("############################################")
                print("### TESTING GenotypeData WITH EACH FILETYPE")
                print("############################################")
                print("\n")

                for ft, aln in testaln.items():
                    if ft == "phylip":
                        data_dir = phylip_files
                    else:
                        data_dir = structure_files

                    alnfile = files(data_dir).joinpath(aln)

                    print("--------------------------------------------------")
                    print(f"--- Testing GenotypeData with {ft} filetype...")
                    print("--------------------------------------------------")
                    print("\n")

                    with as_file(alnfile) as a:
                        data = GenotypeData(
                            filename=a,
                            filetype=ft,
                            popmapfile=m,
                            guidetree=t,
                            qmatrix_iqtree=i,
                            siterates_iqtree=siq,
                        )

                print("-------------------------------------------------------")
                print("--- Testing GenotypeData with non-iqtree rates files...")
                print("-------------------------------------------------------")
                print("\n")

                with as_file(strfile) as a:
                    data = GenotypeData(
                        filename=a,
                        filetype="structure2row",
                        popmapfile=m,
                        guidetree=t,
                        qmatrix=q,
                        siterates=s,
                    )

                    data = GenotypeData(
                        filename=a,
                        filetype="structure2row",
                        popmapfile=m,
                        guidetree=t,
                        qmatrix_iqtree=i,
                        siterates_iqtree=siq,
                    )

                print("++++++++++++++++++++++++++++++++")
                print("+++ SUCCESS!")
                print("++++++++++++++++++++++++++++++++")
                print("\n")

                print("################################")
                print("### TESTING SIMPLE IMPUTERS...")
                print("################################")
                print("\n")

                print("-----------------------------------------------------")
                print("--- Testing ImputeAlleleFreq by-population...")
                print("-----------------------------------------------------")
                print("\n")

                afpops = ImputeAlleleFreq(
                    genotype_data=data,
                    by_populations=True,
                    prefix=prefix,
                    write_output=False,
                )

                print("-----------------------------------------------------")
                print("--- Testing ImputeAlleleFreq global...")
                print("-----------------------------------------------------")
                print("\n")

                afpops = ImputeAlleleFreq(
                    genotype_data=data,
                    by_populations=False,
                    prefix=prefix,
                    write_output=False,
                )

                print("-----------------------------------------------------")
                print("--- Testing ImputePhylo...")
                print("-----------------------------------------------------")
                print("\n")

                phylo = ImputePhylo(
                    genotype_data=data,
                    prefix=prefix,
                    disable_progressbar=True,
                    write_output=False,
                )

                print("-----------------------------------------------------")
                print("--- Testing ImputeNMF...")
                print("-----------------------------------------------------")
                print("\n")

                mf = ImputeNMF(
                    genotype_data=data,
                    prefix=prefix,
                    write_output=False,
                )

                print("++++++++++++++++++++++++++++++++")
                print("+++ SUCCESS!")
                print("++++++++++++++++++++++++++++++++")
                print("\n")

                ##############################################
                ### Make gridparams
                ##############################################

                # For randomizedsearchcv
                # Number of trees in random forest
                n_estimators = [
                    int(x) for x in np.linspace(start=100, stop=1000, num=10)
                ]

                # Number of features to consider at every split
                max_features = ["sqrt", "log2"]

                # Maximum number of levels in the tree
                max_depth = [int(x) for x in np.linspace(10, 110, num=11)]
                max_depth.append(None)

                # Minimmum number of samples required to split a node
                min_samples_split = [int(x) for x in np.linspace(2, 10, num=5)]

                # Minimum number of samples required at each leaf node
                min_samples_leaf = [int(x) for x in np.linspace(1, 5, num=5)]

                # Proportion of dataset to use with bootstrapping
                # max_samples = [x for x in np.linspace(0.5, 1.0, num=6)]

                # Random Forest gridparams - RandomizedSearchCV
                grid_params_random = {
                    "max_features": max_features,
                    "max_depth": max_depth,
                    "min_samples_split": min_samples_split,
                    "min_samples_leaf": min_samples_leaf,
                }

                # Genetic Algorithm grid_params
                grid_params_ga = {
                    "max_features": Categorical(["sqrt", "log2"]),
                    "min_samples_split": Integer(2, 10),
                    "min_samples_leaf": Integer(1, 10),
                    "max_depth": Integer(2, 110),
                }

                print("#################################")
                print("### TESTING IterativeImputer...")
                print("#################################")
                print("\n")

                print("-----------------------------------------------------")
                print(
                    "--- Testing ImputeRandomForest with randomized grid\n"
                    "--- search and initial_strategy == 'populations'..."
                )
                print("-----------------------------------------------------")
                print("\n")

                # Random forest imputation with RandomizedSearchCV grid search
                rf_imp = ImputeRandomForest(
                    data,
                    prefix=prefix,
                    n_estimators=50,
                    n_nearest_features=2,
                    gridparams=grid_params_random,
                    cv=3,
                    grid_iter=40,
                    n_jobs=-1,
                    max_iter=2,
                    column_subset=1.0,
                    ga=False,
                    disable_progressbar=True,
                    extratrees=False,
                    mutation_probability=0.1,
                    chunk_size=1.0,
                    initial_strategy="populations",
                )

                print("-----------------------------------------------------")
                print(
                    "--- Testing ImputeRandomForest with GA grid search and\n"
                    "--- initial_strategy == 'phylogeny'..."
                )
                print("-----------------------------------------------------")
                print("\n")

                # Genetic Algorithm grid search Test
                rf_imp2 = ImputeRandomForest(
                    data,
                    prefix=prefix,
                    n_estimators=50,
                    n_nearest_features=2,
                    gridparams=grid_params_ga,
                    cv=3,
                    grid_iter=40,
                    n_jobs=-1,
                    max_iter=2,
                    column_subset=1.0,
                    ga=True,
                    disable_progressbar=True,
                    extratrees=False,
                    chunk_size=1.0,
                    initial_strategy="phylogeny",
                )

                print("++++++++++++++++++++++++++++++++")
                print("+++ SUCCESS!")
                print("++++++++++++++++++++++++++++++++")
                print("\n")

                print("#################################")
                print("TESTING NEURAL NETWORKS...")
                print("#################################")
                print("\n")

                print("-----------------------------------------------------")
                print(
                    "--- Testing VAE with validation procedure with\n"
                    "--- intial_strategy='populations'..."
                )
                print("-----------------------------------------------------")
                print("\n")

                vae = ImputeVAE(
                    genotype_data=data,
                    prefix=prefix,
                    disable_progressbar=True,
                    validation_only=1.0,
                    initial_strategy="populations",
                    cv=3,
                )

                print("-----------------------------------------------------")
                print(
                    "--- Testing ImputeNLPCA with\n"
                    "--- initial_strategy == 'phylogeny'..."
                )
                print("-----------------------------------------------------")
                print("\n")

                nlpca = ImputeNLPCA(
                    data,
                    n_components=3,
                    initial_strategy="phylogeny",
                    disable_progressbar=True,
                    cv=3,
                    hidden_activation="elu",
                    hidden_layer_sizes="midpoint",
                    validation_only=None,
                    num_hidden_layers=1,
                    learning_rate=0.1,
                )

                print("-------------------------------------------------------")
                print("--- Testing ImputeUBP with initial_strategy == 'nmf'...")
                print("-------------------------------------------------------")
                print("\n")

                ubp = ImputeUBP(
                    genotype_data=data,
                    initial_strategy="nmf",
                    disable_progressbar=True,
                    validation_only=None,
                    learning_rate=0.1,
                    num_hidden_layers=1,
                    hidden_layer_sizes=1,
                    hidden_activation="elu",
                    cv=3,
                    n_components=3,
                )

                print("++++++++++++++++++++++++++++++++")
                print("+++ SUCCESS!")
                print("++++++++++++++++++++++++++++++++")
                print("\n")

                # Try to remove temporary treefile.
                try:
                    os.remove(t)
                except OSError:
                    pass

            print("######################################")
            print("### ALL TESTS PASSED SUCCESSFULLY!")
            print("######################################")
            print("\n")



import numpy as np
import tensorflow as tf

from scikeras.wrappers import KerasClassifier

try:
    from ...utils.scorers import Scorers
    from .models.autoencoder_model import AutoEncoderModel
    from .models.nlpca_model import NLPCAModel
    from .models.ubp_model import UBPPhase1, UBPPhase2, UBPPhase3
    from .models.vae_model import VAEModel
    from ...data_processing.transformers import (
        MLPTargetTransformer,
        UBPInputTransformer,
        AutoEncoderFeatureTransformer,
    )
except (ModuleNotFoundError, ValueError, ImportError):
    from utils.scorers import Scorers
    from impute.unsupervised.neural_network_methods import NeuralNetworkMethods
    from impute.unsupervised.models.vae_model import (
        VAEModel,
    )
    from impute.unsupervised.models.autoencoder_model import AutoEncoderModel
    from impute.unsupervised.models.nlpca_model import NLPCAModel
    from impute.unsupervised.models.ubp_model import (
        UBPPhase1,
        UBPPhase2,
        UBPPhase3,
    )
    from data_processing.transformers import (
        MLPTargetTransformer,
        UBPInputTransformer,
        AutoEncoderFeatureTransformer,
    )


class SAEClassifier(KerasClassifier):
    """Estimator to be used with the scikit-learn API.

    Args:
        output_shape (int): Number of units in model output layer. Defaults to None.

        weights_initializer (str): Kernel initializer to use for model weights. Defaults to "glorot_normal".

        hidden_layer_sizes (List[int]): Output unit size for each hidden layer. Should be list of length num_hidden_layers. Defaults to None.

        num_hidden_layers (int): Number of hidden layers to use. Defaults to 1.

        hidden_activation (str): Hidden activation function to use. Defaults to "elu".

        l1_penalty (float): L1 regularization penalty to use to reduce overfitting. Defautls to 0.01.

        l2_penalty (float): L2 regularization penalty to use to reduce overfitting. Defaults to 0.01.

        dropout_rate (float): Dropout rate for each hidden layer to reduce overfitting. Defaults to 0.2.

        n_components (int): Number of components to use for input V. Defaults to 3.

        num_classes (int, optional): Number of classes in y_train. 012-encoded data should have 3 classes. Defaults to 3.

        kwargs (Any): Other keyword arguments to route to fit, compile, callbacks, etc. Should have the routing prefix (e.g., optimizer__learning_rate=0.01).
    """

    def __init__(
        self,
        y=None,
        output_shape=None,
        weights_initializer="glorot_normal",
        hidden_layer_sizes=None,
        num_hidden_layers=1,
        hidden_activation="elu",
        l1_penalty=0.01,
        l2_penalty=0.01,
        dropout_rate=0.2,
        n_components=3,
        sample_weight=None,
        missing_mask=None,
        num_classes=4,
        activate="sigmoid",
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.y = y
        self.output_shape = output_shape
        self.weights_initializer = weights_initializer
        self.hidden_layer_sizes = hidden_layer_sizes
        self.num_hidden_layers = num_hidden_layers
        self.hidden_activation = hidden_activation
        self.l1_penalty = l1_penalty
        self.l2_penalty = l2_penalty
        self.dropout_rate = dropout_rate
        self.n_components = n_components
        self.sample_weight = sample_weight
        self.missing_mask = missing_mask
        self.num_classes = num_classes
        self.activate = activate

        self.classes_ = np.arange(self.num_classes)
        self.n_classes_ = self.num_classes

    def _keras_build_fn(self, compile_kwargs):
        """Build model with custom parameters.

        Args:
            compile_kwargs (Dict[str, Any]): Dictionary with parameters: values. The parameters should be passed to the class constructor, but should be captured as kwargs. They should also have the routing prefix (e.g., optimizer__learning_rate=0.01). compile_kwargs will automatically be parsed from **kwargs by KerasClassifier and sent here.

        Returns:
            tf.keras.Model: Model instance. The chosen model depends on which phase is passed to the class constructor.
        """

        ######### REMOVING THIS LINE WILL BREAK THE MODEL!!!!! ########
        self.classes_ = np.arange(self.num_classes)

        model = AutoEncoderModel(
            self.y,
            output_shape=self.output_shape,
            n_components=self.n_components,
            weights_initializer=self.weights_initializer,
            hidden_layer_sizes=self.hidden_layer_sizes,
            num_hidden_layers=self.num_hidden_layers,
            hidden_activation=self.hidden_activation,
            l1_penalty=self.l1_penalty,
            l2_penalty=self.l2_penalty,
            dropout_rate=self.dropout_rate,
            sample_weight=self.sample_weight,
            missing_mask=self.missing_mask,
            num_classes=self.num_classes,
        )

        model.compile(
            optimizer=compile_kwargs["optimizer"],
            loss=compile_kwargs["loss"],
            metrics=compile_kwargs["metrics"],
            run_eagerly=False,
        )

        return model

    @staticmethod
    def scorer(y_true, y_pred, **kwargs):
        """Scorer for grid search that masks missing data.

        To use this, do not specify a scoring metric when initializing the grid search object. By default if the scoring_metric option is left as None, then it uses the estimator's scoring metric (this one).

        Args:
            y_true (numpy.ndarray): True target values input to fit().
            y_pred (numpy.ndarray): Predicted target values from estimator. The predictions are modified by self.target_encoder().inverse_transform() before being sent here.
            kwargs (Any): Other parameters sent to sklearn scoring metric. Supported options include missing_mask, scoring_metric, and testing.

        Returns:
            float: Calculated score.
        """
        n_classes_ = kwargs.get("num_classes", 3)
        classes_ = np.arange(n_classes_)
        missing_mask = kwargs.get("missing_mask")

        num_classes = kwargs.get("num_classes", 3)
        testing = kwargs.get("testing", False)

        scorers = Scorers()

        return scorers.scorer(
            y_true,
            y_pred,
            missing_mask=missing_mask,
            num_classes=num_classes,
            testing=testing,
        )

    @property
    def feature_encoder(self):
        """Handles feature input, X, before training.

        Returns:
            MLPTargetTransformer: InputTransformer object that includes fit() and transform() methods to transform input before estimator fitting.
        """
        return AutoEncoderFeatureTransformer(num_classes=self.num_classes)

    @property
    def target_encoder(self):
        """Handles target input and output, y_true and y_pred, both before and after training.

        Returns:
            NNOutputTransformer: NNOutputTransformer object that includes fit(), transform(), and inverse_transform() methods.
        """
        return AutoEncoderFeatureTransformer(
            num_classes=self.num_classes, activate=self.activate
        )

    def predict(self, X, **kwargs):
        """Returns predictions for the given test data.

        Args:
            X (Union[array-like, sparse matrix, dataframe] of shape (n_samples, n_features)): Training samples where n_samples is the number of samples and n_features is the number of features.
        kwargs (Dict[str, Any]): Extra arguments to route to ``Model.predict``\.

        Warnings:
            Passing estimator parameters as keyword arguments (aka as ``**kwargs``) to ``predict`` is not supported by the Scikit-Learn API, and will be removed in a future version of SciKeras. These parameters can also be specified by prefixing ``predict__`` to a parameter at initialization (``BaseWrapper(..., fit__batch_size=32, predict__batch_size=1000)``) or by using ``set_params`` (``est.set_params(fit__batch_size=32, predict__batch_size=1000)``\).

        Returns:
            array-like: Predictions, of shape shape (n_samples,) or (n_samples, n_outputs).

        Notes:
            Had to override predict() here in order to do the __call__ with the refined input, V_latent.
        """
        X_train = self.target_encoder_.transform(X)
        y_pred = self.model_(X_train, training=False)
        return self.target_encoder_.inverse_transform(y_pred)

    def get_metadata(self):
        """Returns a dictionary of meta-parameters generated when this transformer was fitted.

        Used by SciKeras to bind these parameters to the SciKeras estimator itself and make them available as inputs to the Keras model.

        Returns:
            Dict[str, Any]: Dictionary of meta-parameters generated when this transfromer was fitted.
        """
        return {
            "classes_": self.classes_,
            "n_classes_": self.n_classes_,
            "n_outputs_": self.n_outputs_,
            "n_outputs_expected_": self.n_outputs_expected_,
        }


class VAEClassifier(KerasClassifier):
    """Estimator to be used with the scikit-learn API and a keras model.

    Args:
        output_shape (int): Number of units in model output layer. Defaults to None.

        weights_initializer (str, optional): Kernel initializer to use for model weights. Defaults to "glorot_normal".

        hidden_layer_sizes (List[int]): Output unit size for each hidden layer. Should be list of length num_hidden_layers. Defaults to None.

        num_hidden_layers (int, optional): Number of hidden layers to use. Defaults to 1.

        hidden_activation (str, optional): Hidden activation function to use. Defaults to "elu".

        l1_penalty (float, optional): L1 regularization penalty to use to reduce overfitting. Defautls to 0.01.

        l2_penalty (float, optional): L2 regularization penalty to use to reduce overfitting. Defaults to 0.01.

        dropout_rate (float, optional): Dropout rate for each hidden layer to reduce overfitting. Defaults to 0.2.

        kl_beta (float, optional): Kullback-Liebler divergence weight (beta) to apply to KL loss. 1.0 means unweighted, 0.0 means KL loss is not applied at all. Defaults to 1.0.

        n_components (int, optional): Number of components to use for input V. Defaults to 3.

        num_classes (int, optional): Number of classes in y_train. [A,G,C,T...IUPAC codes]-encoded data should have 10 classes. Defaults to 4.

        activate (str or None, optional): If not None, then does the appropriate activation. Multilabel learning uses sigmoid activation, and multiclass uses softmax. If set to None, then the function assumes that the input has already been activated. Possible values include: {None, 'sigmoid', 'softmax'}. Defaults to None.

        kwargs (Any): Other keyword arguments to route to fit, compile, callbacks, etc. Should have the routing prefix (e.g., optimizer__learning_rate=0.01).
    """

    def __init__(
        self,
        output_shape=None,
        weights_initializer="glorot_normal",
        hidden_layer_sizes=None,
        num_hidden_layers=1,
        hidden_activation="elu",
        l1_penalty=0.01,
        l2_penalty=0.01,
        dropout_rate=0.2,
        kl_beta=1.0,
        n_components=3,
        num_classes=4,
        sample_weight=None,
        activate=None,
        y=None,
        missing_mask=None,
        batch_size=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.output_shape = output_shape
        self.weights_initializer = weights_initializer
        self.hidden_layer_sizes = hidden_layer_sizes
        self.num_hidden_layers = num_hidden_layers
        self.hidden_activation = hidden_activation
        self.l1_penalty = l1_penalty
        self.l2_penalty = l2_penalty
        self.dropout_rate = dropout_rate
        self.kl_beta = kl_beta
        self.n_components = n_components
        self.num_classes = num_classes
        self.sample_weight = sample_weight
        self.activate = activate
        self.y = y
        self.missing_mask = missing_mask
        self.batch_size = batch_size

    def _keras_build_fn(self, compile_kwargs):
        """Build model with custom parameters.

        Args:
            compile_kwargs (Dict[str, Any]): Dictionary with parameters: values. The parameters should be passed to the class constructor, but should be captured as kwargs. They should also have the routing prefix (e.g., optimizer__learning_rate=0.01). compile_kwargs will automatically be parsed from **kwargs by KerasClassifier and sent here.

        Returns:
            tf.keras.Model: Model instance. The chosen model depends on which phase is passed to the class constructor.
        """

        ######### REMOVING THIS LINE WILL BREAK THE MODEL!!!!! ########
        self.classes_ = np.arange(self.num_classes)

        model = VAEModel(
            output_shape=self.output_shape,
            n_components=self.n_components,
            weights_initializer=self.weights_initializer,
            hidden_layer_sizes=self.hidden_layer_sizes,
            num_hidden_layers=self.num_hidden_layers,
            hidden_activation=self.hidden_activation,
            l1_penalty=self.l1_penalty,
            l2_penalty=self.l2_penalty,
            dropout_rate=self.dropout_rate,
            kl_beta=self.kl_beta,
            num_classes=self.num_classes,
            sample_weight=self.sample_weight,
            missing_mask=self.missing_mask,
            batch_size=self.batch_size,
            y=self.y,
            final_activation=self.activate,
        )

        model.compile(
            optimizer=compile_kwargs["optimizer"],
            loss=compile_kwargs["loss"],
            metrics=compile_kwargs["metrics"],
            run_eagerly=compile_kwargs["run_eagerly"],
            # sample_weight_mode="temporal",
        )

        return model

    @property
    def feature_encoder(self):
        """Handles feature input, X, before training.

        Returns:
            MLPTargetTransformer: InputTransformer object that includes fit() and transform() methods to transform input before estimator fitting.
        """
        return AutoEncoderFeatureTransformer(num_classes=self.num_classes)

    @property
    def target_encoder(self):
        """Handles target input and output, y_true and y_pred, both before and after training.

        Returns:
            NNOutputTransformer: NNOutputTransformer object that includes fit(), transform(), and inverse_transform() methods.
        """
        return AutoEncoderFeatureTransformer(
            num_classes=self.num_classes,
            activate=self.activate,
        )

    def predict(self, X, **kwargs):
        """Returns predictions for the given test data.

        Args:
            X (Union[array-like, sparse matrix, dataframe] of shape (n_samples, n_features)): Training samples where n_samples is the number of samples and n_features is the number of features.
        kwargs (Dict[str, Any]): Extra arguments to route to ``Model.predict``\.

        Warnings:
            Passing estimator parameters as keyword arguments (aka as ``**kwargs``) to ``predict`` is not supported by the Scikit-Learn API, and will be removed in a future version of SciKeras. These parameters can also be specified by prefixing ``predict__`` to a parameter at initialization (``BaseWrapper(..., fit__batch_size=32, predict__batch_size=1000)``) or by using ``set_params`` (``est.set_params(fit__batch_size=32, predict__batch_size=1000)``\).

        Returns:
            array-like: Predictions, of shape shape (n_samples,) or (n_samples, n_outputs).

        Notes:
            Had to override predict() here in order to do the __call__ with the refined input, V_latent.
        """
        X_train = self.target_encoder_.transform(X)
        y_pred = self.model_(X_train, training=False)
        return self.target_encoder_.inverse_transform(y_pred)

    def get_metadata(self):
        """Returns a dictionary of meta-parameters generated when this transformer was fitted.

        Used by SciKeras to bind these parameters to the SciKeras estimator itself and make them available as inputs to the Keras model.

        Returns:
            Dict[str, Any]: Dictionary of meta-parameters generated when this transfromer was fitted.
        """
        return {
            "classes_": self.classes_,
            "n_classes_": self.n_classes_,
            "n_outputs_": self.n_outputs_,
            "n_outputs_expected_": self.n_outputs_expected_,
        }

    @staticmethod
    def scorer(y_true, y_pred, **kwargs):
        """Scorer for grid search that masks missing data.

        To use this, do not specify a scoring metric when initializing the grid search object. By default if the scoring_metric option is left as None, then it uses the estimator's scoring metric (this one).

        Args:
            y_true (numpy.ndarray): True target values input to fit().

            y_pred (numpy.ndarray): Predicted target values from estimator. The predictions are modified by self.target_encoder().inverse_transform() before being sent here.

            kwargs (Any): Other parameters sent to sklearn scoring metric. Supported options include missing_mask, scoring_metric, and testing.

        Returns:
            float: Calculated score.
        """

        n_classes_ = kwargs.get("num_classes", 3)
        classes_ = np.arange(n_classes_)
        missing_mask = kwargs.get("missing_mask")

        num_classes = kwargs.get("num_classes", 3)
        testing = kwargs.get("testing", False)

        y_pred = y_pred.reshape(y_pred.shape[0], -1, num_classes)

        scorers = Scorers()

        return scorers.scorer(
            y_true,
            y_pred,
            missing_mask=missing_mask,
            num_classes=num_classes,
            testing=testing,
        )


class MLPClassifier(KerasClassifier):
    """Estimator to be used with the scikit-learn API.

    Args:
        V (numpy.ndarray or Dict[str, Any]): Input X values of shape (n_samples, n_components). If a dictionary is passed, each key: value pair should have randomly initialized values for n_components: V. self.feature_encoder() will parse it and select the key: value pair with the current n_components. This allows n_components to be grid searched using GridSearchCV. Otherwise, it throws an error that the dimensions are off. Defaults to None.

        y_train (numpy.ndarray): One-hot encoded target data. Defaults to None.

        ubp_weights (tensorflow.Tensor): Weights from UBP model. Fetched by doing model.get_weights() on phase 2 model. Only used if phase 3. Defaults to None.

        batch_size (int): Batch size to train with. Defaults to 32.

        missing_mask (np.ndarray): Missing mask with missing values set to False (0) and observed values as True (1). Defaults to None. Defaults to None.

        output_shape (int): Number of units in model output layer. Defaults to None.

        weights_initializer (str): Kernel initializer to use for model weights. Defaults to "glorot_normal".

        hidden_layer_sizes (List[int]): Output unit size for each hidden layer. Should be list of length num_hidden_layers. Defaults to None.

        num_hidden_layers (int): Number of hidden layers to use. Defaults to 1.

        hidden_activation (str): Hidden activation function to use. Defaults to "elu".

        l1_penalty (float): L1 regularization penalty to use to reduce overfitting. Defautls to 0.01.

        l2_penalty (float): L2 regularization penalty to use to reduce overfitting. Defaults to 0.01.

        dropout_rate (float): Dropout rate for each hidden layer to reduce overfitting. Defaults to 0.2.

        num_classes (int): Number of classes in output predictions. Defaults to 3.

        phase (int or None): Current phase (if doing UBP), or None if doing NLPCA. Defults to None.

        sample_weight (numpy.ndarray): Sample weight matrix for reducing the impact of class imbalance. Should be of shape (n_samples, n_features).

        n_components (int): Number of components to use for input V. Defaults to 3.

        kwargs (Any): Other keyword arguments to route to fit, compile, callbacks, etc. Should have the routing prefix (e.g., optimizer__learning_rate=0.01).
    """

    def __init__(
        self,
        V,
        y_train,
        ubp_weights=None,
        batch_size=32,
        missing_mask=None,
        output_shape=None,
        weights_initializer="glorot_normal",
        hidden_layer_sizes=None,
        num_hidden_layers=1,
        hidden_activation="elu",
        l1_penalty=0.01,
        l2_penalty=0.01,
        dropout_rate=0.2,
        num_classes=3,
        phase=None,
        sample_weight=None,
        n_components=3,
        activate=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.V = V
        self.y_train = y_train
        self.ubp_weights = ubp_weights
        self.batch_size = batch_size
        self.missing_mask = missing_mask
        self.output_shape = output_shape
        self.weights_initializer = weights_initializer
        self.hidden_layer_sizes = hidden_layer_sizes
        self.num_hidden_layers = num_hidden_layers
        self.hidden_activation = hidden_activation
        self.l1_penalty = l1_penalty
        self.l2_penalty = l2_penalty
        self.dropout_rate = dropout_rate
        self.num_classes = num_classes
        self.phase = phase
        self.sample_weight = sample_weight
        self.n_components = n_components
        self.activate = activate

    def _keras_build_fn(self, compile_kwargs):
        """Build model with custom parameters.

        Args:
            compile_kwargs (Dict[str, Any]): Dictionary with parameters: values. The parameters should be passed to the class constructor, but should be captured as kwargs. They should also have the routing prefix (e.g., optimizer__learning_rate=0.01). compile_kwargs will automatically be parsed from **kwargs by KerasClassifier and sent here.

        Returns:
            tf.keras.Model: Model instance. The chosen model depends on which phase is passed to the class constructor.
        """
        ######### REMOVING THIS LINE WILL BREAK THE MODEL!!!!! ########
        self.classes_ = np.arange(self.num_classes)

        if self.phase is None:
            model = NLPCAModel(
                V=self.V,
                y=self.y_train,
                batch_size=self.batch_size,
                missing_mask=self.missing_mask,
                output_shape=self.output_shape,
                n_components=self.n_components,
                weights_initializer=self.weights_initializer,
                hidden_layer_sizes=self.hidden_layer_sizes,
                num_hidden_layers=self.num_hidden_layers,
                hidden_activation=self.hidden_activation,
                l1_penalty=self.l1_penalty,
                l2_penalty=self.l2_penalty,
                dropout_rate=self.dropout_rate,
                num_classes=self.num_classes,
                phase=self.phase,
                sample_weight=self.sample_weight,
            )

        elif self.phase == 1:
            model = UBPPhase1(
                V=self.V,
                y=self.y_train,
                batch_size=self.batch_size,
                missing_mask=self.missing_mask,
                output_shape=self.output_shape,
                n_components=self.n_components,
                weights_initializer=self.weights_initializer,
                hidden_layer_sizes=self.hidden_layer_sizes,
                num_hidden_layers=self.num_hidden_layers,
                l1_penalty=self.l1_penalty,
                l2_penalty=self.l2_penalty,
                dropout_rate=self.dropout_rate,
                num_classes=self.num_classes,
                phase=self.phase,
            )

        elif self.phase == 2:
            model = UBPPhase2(
                V=self.V,
                y=self.y_train,
                batch_size=self.batch_size,
                missing_mask=self.missing_mask,
                output_shape=self.output_shape,
                n_components=self.n_components,
                weights_initializer=self.weights_initializer,
                hidden_layer_sizes=self.hidden_layer_sizes,
                num_hidden_layers=self.num_hidden_layers,
                hidden_activation=self.hidden_activation,
                l1_penalty=self.l1_penalty,
                l2_penalty=self.l2_penalty,
                dropout_rate=self.dropout_rate,
                num_classes=self.num_classes,
                phase=self.phase,
            )

        elif self.phase == 3:
            model = UBPPhase3(
                V=self.V,
                y=self.y_train,
                batch_size=self.batch_size,
                missing_mask=self.missing_mask,
                output_shape=self.output_shape,
                n_components=self.n_components,
                weights_initializer=self.weights_initializer,
                hidden_layer_sizes=self.hidden_layer_sizes,
                num_hidden_layers=self.num_hidden_layers,
                hidden_activation=self.hidden_activation,
                dropout_rate=self.dropout_rate,
                num_classes=self.num_classes,
                phase=self.phase,
            )

            model.build((None, self.n_components))

        model.compile(
            optimizer=compile_kwargs["optimizer"],
            loss=compile_kwargs["loss"],
            metrics=compile_kwargs["metrics"],
            run_eagerly=True,
        )

        model.set_model_outputs()

        if self.phase == 3:
            model.set_weights(self.ubp_weights)

        return model

    @staticmethod
    def scorer(y_true, y_pred, **kwargs):
        """Scorer for grid search that masks missing data.

        To use this, do not specify a scoring metric when initializing the grid search object. By default if the scoring_metric option is left as None, then it uses the estimator's scoring metric (this one).

        Args:
            y_true (numpy.ndarray): True target values input to fit().

            y_pred (numpy.ndarray): Predicted target values from estimator. The predictions are modified by self.target_encoder().inverse_transform() before being sent here.

            kwargs (Any): Other parameters sent to sklearn scoring metric. Supported options include missing_mask, scoring_metric, and testing.

        Returns:
            float: Calculated score.
        """
        missing_mask = kwargs.get(
            "missing_mask", np.ones(y_true.shape, dtype=bool)
        )
        num_classes = kwargs.get("num_classes", 3)
        testing = kwargs.get("testing", False)

        scorers = Scorers()

        return scorers.scorer(
            y_true,
            y_pred,
            missing_mask=missing_mask,
            num_classes=num_classes,
            testing=testing,
        )

    @property
    def feature_encoder(self):
        """Handles feature input, X, before training.

        Returns:
            UBPInputTransformer: InputTransformer object that includes fit() and transform() methods to transform input before estimator fitting.
        """
        return UBPInputTransformer(self.n_components, self.V)

    @property
    def target_encoder(self):
        """Handles target input and output, y_true and y_pred, both before and after training.

        Returns:
            NNOutputTransformer: NNOutputTransformer object that includes fit(), transform(), and inverse_transform() methods.
        """
        return AutoEncoderFeatureTransformer(
            num_classes=self.num_classes,
            activate=None,
        )

    def predict(self, X, **kwargs):
        """Returns predictions for the given test data.

        Args:
            X (Union[array-like, sparse matrix, dataframe] of shape (n_samples, n_features)): Training samples where n_samples is the number of samples and n_features is the number of features.
        kwargs (Dict[str, Any]): Extra arguments to route to ``Model.predict``\.

        Warnings:
            Passing estimator parameters as keyword arguments (aka as ``**kwargs``) to ``predict`` is not supported by the Scikit-Learn API, and will be removed in a future version of SciKeras. These parameters can also be specified by prefixing ``predict__`` to a parameter at initialization (``BaseWrapper(..., fit__batch_size=32, predict__batch_size=1000)``) or by using ``set_params`` (``est.set_params(fit__batch_size=32, predict__batch_size=1000)``\).

        Returns:
            array-like: Predictions, of shape shape (n_samples,) or (n_samples, n_outputs).

        Notes:
            Had to override predict() here in order to do the __call__ with the refined input, V_latent.
        """
        y_pred_proba = self.model_(self.model_.V_latent, training=False)
        return self.target_encoder_.inverse_transform(y_pred_proba)

    def get_metadata(self):
        """Returns a dictionary of meta-parameters generated when this transformer was fitted.

        Used by SciKeras to bind these parameters to the SciKeras estimator itself and make them available as inputs to the Keras model.

        Returns:
            Dict[str, Any]: Dictionary of meta-parameters generated when this transfromer was fitted.
        """
        return {
            "classes_": self.classes_,
            "n_classes_": self.n_classes_,
            "n_outputs_": self.n_outputs_,
            "n_outputs_expected_": self.n_outputs_expected_,
        }



import copy
import os
import logging
import sys
import warnings

import numpy as np
import pandas as pd

# Third-party imports
import numpy as np
import pandas as pd

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.impute import SimpleImputer
from sklearn.metrics import (
    roc_auc_score,
    precision_recall_fscore_support,
    average_precision_score,
)
from sklearn.preprocessing import label_binarize

# Import tensorflow with reduced warnings.
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
logging.getLogger("tensorflow").disabled = True
warnings.filterwarnings("ignore", category=UserWarning)

# noinspection PyPackageRequirements
import tensorflow as tf

# Disable can't find cuda .dll errors. Also turns of GPU support.
tf.config.set_visible_devices([], "GPU")

from tensorflow.python.util import deprecation

# Disable warnings and info logs.
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
tf.get_logger().setLevel(logging.ERROR)


# Monkey patching deprecation utils to supress warnings.
# noinspection PyUnusedLocal
def deprecated(
    date, instructions, warn_once=True
):  # pylint: disable=unused-argument
    def deprecated_wrapper(func):
        return func

    return deprecated_wrapper


deprecation.deprecated = deprecated

# Custom Modules
try:
    from ..utils import misc

except (ModuleNotFoundError, ValueError, ImportError):
    from pgsui.utils import misc


# Pandas on pip gives a performance warning when doing the below code.
# Apparently it's a bug that exists in the pandas version I used here.
# It can be safely ignored.
warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)


def encode_onehot(X):
    """Convert 012-encoded data to one-hot encodings.
    Args:
        X (numpy.ndarray): Input array with 012-encoded data and -9 as the missing data value.
    Returns:
        pandas.DataFrame: One-hot encoded data, ignoring missing values (np.nan).
    """
    Xt = np.zeros(shape=(X.shape[0], X.shape[1], 3))
    mappings = {
        0: np.array([1, 0, 0]),
        1: np.array([0, 1, 0]),
        2: np.array([0, 0, 1]),
        -9: np.array([np.nan, np.nan, np.nan]),
    }
    for row in np.arange(X.shape[0]):
        Xt[row] = [mappings[enc] for enc in X[row]]
    return Xt


def mle(row):
    """Get the Maximum Likelihood Estimation for the best prediction. Basically, it sets the index of the maxiumum value in a vector (row) to 1.0, since it is one-hot encoded.

    Args:
        row (numpy.ndarray(float)): Row vector with predicted values as floating points.

    Returns:
        numpy.ndarray(float): Row vector with the highest prediction set to 1.0 and the others set to 0.0.
    """
    res = np.zeros(row.shape[0])
    res[np.argmax(row)] = 1
    return res


class UBPInputTransformer(BaseEstimator, TransformerMixin):
    """Transform input X prior to estimator fitting.

    Args:
        n_components (int): Number of principal components currently being used in V.

        V (numpy.ndarray or Dict[str, Any]): If doing grid search, should be a dictionary with current_component: numpy.ndarray. If not doing grid search, then it should be a numpy.ndarray.
    """

    def __init__(self, n_components, V):
        self.n_components = n_components
        self.V = V

    def fit(self, X):
        """Fit transformer to input data X.

        Args:
            X (numpy.ndarray): Input data to fit. If numpy.ndarray, then should be of shape (n_samples, n_components). If dictionary, then should be component: numpy.ndarray.

        Returns:
            self: Class instance.
        """
        self.n_features_in_ = self.n_components
        return self

    def transform(self, X):
        """Transform input data X to the needed format.

        Args:
            X (numpy.ndarray): Input data to fit. If numpy.ndarray, then should be of shape (n_samples, n_components). If dictionary, then should be component: numpy.ndarray.

        Returns:
            numpy.ndarray: Formatted input data with correct component.

        Raises:
            TypeError: V must be a dictionary if phase is None or phase == 1.
            TypeError: V must be a numpy array if phase is 2 or 3.
        """
        if not isinstance(self.V, dict):
            raise TypeError(f"V must be a dictionary, but got {type(self.V)}")
        return self.V[self.n_components]


class AutoEncoderFeatureTransformer(BaseEstimator, TransformerMixin):
    """Transformer to format autoencoder features and targets before model fitting.

    The input data, X, is encoded to one-hot format, and then missing values are filled to [-1] * num_classes.

    Missing and observed boolean masks are also generated.

    Args:
        num_classes (int, optional): The number of classes in the last axis dimention of the input array. Defaults to 3.

        return_int (bool, optional): Whether to return an integer-encoded array (If True) or a one-hot or multi-label encoded array (If False.). Defaults to False.

        activate (str or None, optional): If not None, then does the appropriate activation. Multilabel learning uses sigmoid activation, and multiclass uses softmax. If set to None, then the function assumes that the input has already been activated. Possible values include: {None, 'sigmoid', 'softmax'}. Defaults to None.
    """

    def __init__(self, num_classes=3, return_int=False, activate=None):
        self.num_classes = num_classes
        self.return_int = return_int
        self.activate = activate

    def fit(self, X, y=None):
        """set attributes used to transform X (input features).

        Args:
            X (numpy.ndarray): Input integer-encoded numpy array.

            y (None): Just for compatibility with sklearn API.
        """
        X = misc.validate_input_type(X, return_type="array")

        self.X_decoded = X

        # VAE uses 4 classes ([A,T,G,C]), SAE uses 3 ([0,1,2]).
        if self.num_classes == 3:
            enc_func = self.encode_012
        elif self.num_classes == 4:
            enc_func = self.encode_multilab
        elif self.num_classes == 10:
            enc_func = self.encode_multiclass
        else:
            raise ValueError(
                f"Invalid value passed to num_classes in "
                f"AutoEncoderFeatureTransformer. Only 3 or 4 are supported, "
                f"but got {self.num_classes}."
            )

        # Encode the data.
        self.X_train = enc_func(X)
        self.classes_ = np.arange(self.num_classes)
        self.n_classes_ = self.num_classes

        # Get missing and observed data boolean masks.
        self.missing_mask_, self.observed_mask_ = self._get_masks(self.X_train)

        # To accomodate multiclass-multioutput.
        self.n_outputs_expected_ = 1

        self.n_outputs_ = self.X_train.shape[1]

        return self

    def transform(self, X):
        """Transform X to one-hot encoded format.

        Accomodates multiclass targets with a 3D shape.

        Args:
            X (numpy.ndarray): One-hot encoded target data of shape (n_samples, n_features, num_classes).

        Returns:
            numpy.ndarray: Transformed target data in one-hot format of shape (n_samples, n_features, num_classes).
        """
        if self.return_int:
            return X
        else:
            # X = misc.validate_input_type(X, return_type="array")
            return self._fill(self.X_train, self.missing_mask_)

    def inverse_transform(self, y, return_proba=False):
        """Transform target to output format.

        Args:
            y (numpy.ndarray): Array to inverse transform.

            return_proba (bool): Just for compatibility with scikeras API.
        """
        try:
            if self.activate is None:
                y = y.numpy()
            elif self.activate == "softmax":
                y = tf.nn.softmax(y).numpy()
            elif self.activate == "sigmoid":
                y = tf.nn.sigmoid(y).numpy()
            else:
                raise ValueError(
                    f"Invalid value passed to keyword argument activate. Valid "
                    f"options include: None, 'softmax', or 'sigmoid', but got "
                    f"{self.activate}"
                )
        except AttributeError:
            # If numpy array already.
            if self.activate is None:
                y = y.copy()
            elif self.activate == "softmax":
                y = tf.nn.softmax(tf.convert_to_tensor(y)).numpy()
            elif self.activate == "sigmoid":
                y = tf.nn.sigmoid(tf.convert_to_tensor(y)).numpy()
            else:
                raise ValueError(
                    f"Invalid value passed to keyword argument activate. Valid "
                    f"options include: None, 'softmax', or 'sigmoid', but got "
                    f"{self.activate}"
                )
        return y

    def encode_012(self, X):
        """Convert 012-encoded data to one-hot encodings.
        Args:
            X (numpy.ndarray): Input array with 012-encoded data and -9 as the missing data value.
        Returns:
            pandas.DataFrame: One-hot encoded data, ignoring missing values (np.nan).
        """
        Xt = np.zeros(shape=(X.shape[0], X.shape[1], 3))
        mappings = {
            0: np.array([1, 0, 0]),
            1: np.array([0, 1, 0]),
            2: np.array([0, 0, 1]),
            -9: np.array([np.nan, np.nan, np.nan]),
        }
        for row in np.arange(X.shape[0]):
            Xt[row] = [mappings[enc] for enc in X[row]]
        return Xt

    def encode_multilab(self, X, multilab_value=1.0):
        """Encode 0-9 integer data in multi-label one-hot format.
        Args:
            X (numpy.ndarray): Input array with 012-encoded data and -9 as the missing data value.

            multilab_value (float): Value to use for multilabel target encodings. Defaults to 0.5.
        Returns:
            pandas.DataFrame: One-hot encoded data, ignoring missing values (np.nan). multi-label categories will be encoded as 0.5. Otherwise, it will be 1.0.
        """
        Xt = np.zeros(shape=(X.shape[0], X.shape[1], 4))
        mappings = {
            0: [1.0, 0.0, 0.0, 0.0],
            1: [0.0, 1.0, 0.0, 0.0],
            2: [0.0, 0.0, 1.0, 0.0],
            3: [0.0, 0.0, 0.0, 1.0],
            4: [multilab_value, multilab_value, 0.0, 0.0],
            5: [multilab_value, 0.0, multilab_value, 0.0],
            6: [multilab_value, 0.0, 0.0, multilab_value],
            7: [0.0, multilab_value, multilab_value, 0.0],
            8: [0.0, multilab_value, 0.0, multilab_value],
            9: [0.0, 0.0, multilab_value, multilab_value],
            -9: [np.nan, np.nan, np.nan, np.nan],
        }
        for row in np.arange(X.shape[0]):
            Xt[row] = [mappings[enc] for enc in X[row]]
        return Xt

    def decode_multilab(self, X, multilab_value=1.0):
        """Decode one-hot format data back to 0-9 integer data.

        Args:
            X (numpy.ndarray): Input array with one-hot-encoded data.

            multilab_value (float): Value to use for multilabel target encodings. Defaults to 0.5.

        Returns:
            pandas.DataFrame: Decoded data, with multi-label categories decoded to their original integer representation.
        """
        Xt = np.zeros(shape=(X.shape[0], X.shape[1]))
        mappings = {
            tuple([1.0, 0.0, 0.0, 0.0]): 0,
            tuple([0.0, 1.0, 0.0, 0.0]): 1,
            tuple([0.0, 0.0, 1.0, 0.0]): 2,
            tuple([0.0, 0.0, 0.0, 1.0]): 3,
            tuple([multilab_value, multilab_value, 0.0, 0.0]): 4,
            tuple([multilab_value, 0.0, multilab_value, 0.0]): 5,
            tuple([multilab_value, 0.0, 0.0, multilab_value]): 6,
            tuple([0.0, multilab_value, multilab_value, 0.0]): 7,
            tuple([0.0, multilab_value, 0.0, multilab_value]): 8,
            tuple([0.0, 0.0, multilab_value, multilab_value]): 9,
            tuple([np.nan, np.nan, np.nan, np.nan]): -9,
        }
        for row in np.arange(X.shape[0]):
            Xt[row] = [mappings[tuple(enc)] for enc in X[row]]
        return Xt

    def encode_multiclass(self, X, num_classes=10, missing_value=-9):
        """Encode 0-9 integer data in multi-class one-hot format.

        Missing values get encoded as ``[np.nan] * num_classes``
        Args:
            X (numpy.ndarray): Input array with 012-encoded data and ``missing_value`` as the missing data value.

            num_classes (int, optional): Number of classes to use. Defaults to 10.

            missing_value (int, optional): Missing data value to replace with ``[np.nan] * num_classes``\. Defaults to -9.
        Returns:
            pandas.DataFrame: Multi-class one-hot encoded data, ignoring missing values (np.nan).
        """
        int_cats, ohe_arr = np.arange(num_classes), np.eye(num_classes)
        mappings = dict(zip(int_cats, ohe_arr))
        mappings[missing_value] = np.array([np.nan] * num_classes)

        Xt = np.zeros(shape=(X.shape[0], X.shape[1], num_classes))
        for row in np.arange(X.shape[0]):
            Xt[row] = [mappings[enc] for enc in X[row]]
        return Xt

    def _fill(self, data, missing_mask, missing_value=-1):
        """Mask missing data as ``missing_value``\.

        Args:
            data (numpy.ndarray): Input with missing values of shape (n_samples, n_features, num_classes).

            missing_mask (np.ndarray(bool)): Missing data mask with True corresponding to a missing value.

            missing_value (int): Value to set missing data to. If a list is provided, then its length should equal the number of one-hot classes.
        """
        if self.num_classes > 1:
            missing_value = [missing_value] * self.num_classes
        data[missing_mask] = missing_value
        return data

    def _get_masks(self, X):
        """Format the provided target data for use with UBP/NLPCA.

        Args:
            y (numpy.ndarray(float)): Input data that will be used as the target of shape (n_samples, n_features, num_classes).

        Returns:
            numpy.ndarray(float): Missing data mask, with missing values encoded as 1's and non-missing as 0's.

            numpy.ndarray(float): Observed data mask, with non-missing values encoded as 1's and missing values as 0's.
        """
        missing_mask = self._create_missing_mask(X)
        observed_mask = ~missing_mask
        return missing_mask, observed_mask

    def _create_missing_mask(self, data):
        """Creates a missing data mask with boolean values.
        Args:
            data (numpy.ndarray): Data to generate missing mask from, of shape (n_samples, n_features, n_classes).
        Returns:
            numpy.ndarray(bool): Boolean mask of missing values of shape (n_samples, n_features), with True corresponding to a missing data point.
        """
        return np.isnan(data).all(axis=2)


class MLPTargetTransformer(BaseEstimator, TransformerMixin):
    """Transformer to format UBP / NLPCA target data both before and after model fitting."""

    def fit(self, y):
        """Fit 012-encoded target data.

        Args:
            y (numpy.ndarray): Target data that is 012-encoded.

        Returns:
            self: Class instance.
        """
        y = misc.validate_input_type(y, return_type="array")

        # Original 012-encoded y
        self.y_decoded_ = y

        y_train = encode_onehot(y)

        # Get missing and observed data boolean masks.
        self.missing_mask_, self.observed_mask_ = self._get_masks(y_train)

        # To accomodate multiclass-multioutput.
        self.n_outputs_expected_ = 1

        return self

    def transform(self, y):
        """Transform y_true to one-hot encoded.

        Accomodates multiclass-multioutput targets.

        Args:
            y (numpy.ndarray): One-hot encoded target data.

        Returns:
            numpy.ndarray: y_true target data.
        """
        y = misc.validate_input_type(y, return_type="array")
        y_train = encode_onehot(y)
        return self._fill(y_train, self.missing_mask_)

    def inverse_transform(self, y):
        """Decode y_pred from one-hot to 012-based encoding.

        This allows sklearn.metrics to be used.

        Args:
            y (numpy.ndarray): One-hot encoded predicted probabilities after model fitting.

        Returns:
            numpy.ndarray: y predictions in same format as y_true.
        """
        # VAE has tuple output
        if isinstance(y, tuple):
            y = y[0]

        # Return predictions.
        return tf.nn.softmax(y).numpy()

    def _fill(self, data, missing_mask, missing_value=-1, num_classes=3):
        """Mask missing data as ``missing_value``\.

        Args:
            data (numpy.ndarray): Input with missing values of shape (n_samples, n_features, num_classes).

            missing_mask (np.ndarray(bool)): Missing data mask with True corresponding to a missing value.

            missing_value (int): Value to set missing data to. If a list is provided, then its length should equal the number of one-hot classes. Defaults to -1.

            num_classes (int): Number of classes in dataset. Defaults to 3.
        """
        if num_classes > 1:
            missing_value = [missing_value] * num_classes
        data[missing_mask] = missing_value
        return data

    def _get_masks(self, X):
        """Format the provided target data for use with UBP/NLPCA.

        Args:
            X (numpy.ndarray(float)): Input data that will be used as the target.

        Returns:
            numpy.ndarray(float): Missing data mask, with missing values encoded as 1's and non-missing as 0's.

            numpy.ndarray(float): Observed data mask, with non-missing values encoded as 1's and missing values as 0's.
        """
        missing_mask = self._create_missing_mask(X)
        observed_mask = ~missing_mask
        return missing_mask, observed_mask

    def _create_missing_mask(self, data):
        """Creates a missing data mask with boolean values.
        Args:
            data (numpy.ndarray): Data to generate missing mask from, of shape (n_samples, n_features, n_classes).
        Returns:
            numpy.ndarray(bool): Boolean mask of missing values of shape (n_samples, n_features), with True corresponding to a missing data point.
        """
        return np.isnan(data).all(axis=2)

    def _decode(self, y):
        """Evaluate UBP / NLPCA predictions by calculating the highest predicted value.

        Calucalates highest predicted value for each row vector and each class, setting the most likely class to 1.0.

        Args:
            y (numpy.ndarray): Input one-hot encoded data.

        Returns:
            numpy.ndarray: Imputed one-hot encoded values.
        """
        Xprob = y
        Xt = np.apply_along_axis(mle, axis=2, arr=Xprob)
        Xpred = np.argmax(Xt, axis=2)
        Xtrue = np.argmax(y, axis=2)
        Xdecoded = np.zeros((Xpred.shape[0], Xpred.shape[1]))
        for idx in np.arange(Xdecoded):
            imputed_idx = np.where(self.observed_mask_[idx] == 0)
            known_idx = np.nonzero(self.observed_mask_[idx])
            Xdecoded[idx, imputed_idx] = Xpred[idx, imputed_idx]
            Xdecoded[idx, known_idx] = Xtrue[idx, known_idx]
        return Xdecoded.astype("int8")


class UBPTargetTransformer(BaseEstimator, TransformerMixin):
    """Transformer to format UBP / NLPCA target data both before model fitting.

    Examples:
        >>>ubp_tt = UBPTargetTransformer()
        >>>y_train = ubp_tt.fit_transform(y)
    """

    def fit(self, y):
        """Fit 012-encoded target data.

        Args:
            y (numpy.ndarray): Target data that is 012-encoded, of shape (n_samples, n_features).

        Returns:
            self: Class instance.
        """
        y = misc.validate_input_type(y, return_type="array")

        # Original 012-encoded y
        self.y_decoded_ = y

        # One-hot encode y.
        y_train = encode_onehot(y)

        # Get missing and observed data boolean masks.
        self.missing_mask_, self.observed_mask_ = self._get_masks(y_train)

        # To accomodate multiclass-multioutput.
        self.n_outputs_expected_ = 1

        return self

    def transform(self, y):
        """Transform 012-encoded target to one-hot encoded format.

        Accomodates multiclass-multioutput targets.

        Args:
            y (numpy.ndarray): One-hot encoded target data of shape (n_samples, n_features).

        Returns:
            numpy.ndarray: y_true target data.
        """
        y = misc.validate_input_type(y, return_type="array")
        y_train = encode_onehot(y)
        return self._fill(y_train, self.missing_mask_)

    def inverse_transform(self, y):
        """Decode y_predicted from one-hot to 012-integer encoding.

        Performs a softmax activation for multiclass classification.

        This allows sklearn.metrics to be used.

        Args:
            y (numpy.ndarray): One-hot encoded predicted probabilities after model fitting, of shape (n_samples, n_features, num_classes).

        Returns:
            numpy.ndarray: y predictions in same format as y_true (n_samples, n_features).
        """
        return tf.nn.softmax(y).numpy()

    def _fill(self, data, missing_mask, missing_value=-1, num_classes=3):
        """Mask missing data as ``missing_value``\.

        Args:
            data (numpy.ndarray): Input with missing values of shape (n_samples, n_features, num_classes).

            missing_mask (np.ndarray(bool)): Missing data mask with True corresponding to a missing value, of shape (n_samples, n_features).

            missing_value (int, optional): Value to set missing data to. If a list is provided, then its length should equal the number of one-hot classes. Defaults to -1.

            num_classes (int, optional): Number of classes to use. Defaults to 3.
        """
        if num_classes > 1:
            missing_value = [missing_value] * num_classes
        data[missing_mask] = missing_value
        return data

    def _get_masks(self, y):
        """Format the provided target data for use with UBP/NLPCA models.

        Args:
            y (numpy.ndarray(float)): Input data that will be used as the target of shape (n_samples, n_features, num_classes).

        Returns:
            numpy.ndarray(float): Missing data mask, with missing values encoded as 1's and non-missing as 0's.

            numpy.ndarray(float): Observed data mask, with non-missing values encoded as 1's and missing values as 0's.
        """
        missing_mask = self._create_missing_mask(y)
        observed_mask = ~missing_mask
        return missing_mask, observed_mask

    def _create_missing_mask(self, data):
        """Creates a missing data mask with boolean values.

        Args:
            data (numpy.ndarray): Data to generate missing mask from, of shape (n_samples, n_features, n_classes).

        Returns:
            numpy.ndarray(bool): Boolean mask of missing values of shape (n_samples, n_features), with True corresponding to a missing data point.
        """
        return np.isnan(data).all(axis=2)

    def _decode(self, y):
        """Evaluate UBP/NLPCA predictions by calculating the argmax.

        Calucalates highest predicted value for each row vector and each class, setting the most likely class to 1.0.

        Args:
            y (numpy.ndarray): Input one-hot encoded data of shape (n_samples, n_features, num_classes).

        Returns:
            numpy.ndarray: Imputed one-hot encoded values.
        """
        Xprob = y
        Xt = np.apply_along_axis(mle, axis=2, arr=Xprob)
        Xpred = np.argmax(Xt, axis=2)
        Xtrue = np.argmax(y, axis=2)
        Xdecoded = np.zeros((Xpred.shape[0], Xpred.shape[1]))
        for idx in np.arange(Xdecoded):
            imputed_idx = np.where(self.observed_mask_[idx] == 0)
            known_idx = np.nonzero(self.observed_mask_[idx])
            Xdecoded[idx, imputed_idx] = Xpred[idx, imputed_idx]
            Xdecoded[idx, known_idx] = Xtrue[idx, known_idx]
        return Xdecoded.astype("int8")


class SimGenotypeDataTransformer(BaseEstimator, TransformerMixin):
    """Simulate missing data on genotypes read/ encoded in a GenotypeData object.

    Copies metadata from a GenotypeData object and simulates user-specified proportion of missing data

    Args:
        genotype_data (GenotypeData object): GenotypeData instance.

        prop_missing (float, optional): Proportion of missing data desired in output. Defaults to 0.1

        strategy (str, optional): Strategy for simulating missing data. May be one of: "nonrandom", "nonrandom_weighted", "random_weighted", "random_weighted_inv", or "random". When set to "nonrandom", branches from GenotypeData.guidetree will be randomly sampled to generate missing data on descendant nodes. For "nonrandom_weighted", missing data will be placed on nodes proportionally to their branch lengths (e.g., to generate data distributed as might be the case with mutation-disruption of RAD sites). Defaults to "random"

        missing_val (int, optional): Value that represents missing data. Defaults to -9.

        mask_missing (bool, optional): True if you want to skip original missing values when simulating new missing data, False otherwise. Defaults to True.

        verbose (bool, optional): Verbosity level. Defaults to 0.

        tol (float): Tolerance to reach proportion specified in self.prop_missing. Defaults to 1/num_snps*num_inds

        max_tries (int): Maximum number of tries to reach targeted missing data proportion within specified tol. If None, num_inds will be used. Defaults to None.

    Attributes:

        original_missing_mask_ (numpy.ndarray): Array with boolean mask for original missing locations.

        simulated_missing_mask_ (numpy.ndarray): Array with boolean mask for simulated missing locations, excluding the original ones.

        all_missing_mask_ (numpy.ndarray): Array with boolean mask for all missing locations, including both simulated and original.

    Properties:
        missing_count (int): Number of genotypes masked by chosen missing data strategy

        prop_missing_real (float): True proportion of missing data generated using chosen strategy

        mask (numpy.ndarray): 2-dimensional array tracking the indices of sampled missing data sites (n_samples, n_sites)
    """

    def __init__(
        self,
        genotype_data,
        *,
        prop_missing=0.1,
        strategy="random",
        missing_val=-9,
        mask_missing=True,
        verbose=0,
        tol=None,
        max_tries=None,
    ) -> None:
        self.genotype_data = genotype_data
        self.prop_missing = prop_missing
        self.strategy = strategy
        self.missing_val = missing_val
        self.mask_missing = mask_missing
        self.verbose = verbose
        self.tol = tol
        self.max_tries = max_tries

    def fit(self, X):
        """Fit to input data X by simulating missing data.

        Missing data will be simulated in varying ways depending on the ``strategy`` setting.

        Args:
            X (pandas.DataFrame, numpy.ndarray, or List[List[int]]): Data with which to simulate missing data. It should have already been imputed with one of the non-machine learning simple imputers, and there should be no missing data present in X.

        Raises:
            TypeError: SimGenotypeData.tree must not be NoneType when using strategy="nonrandom" or "nonrandom_weighted".

            ValueError: Invalid ``strategy`` parameter provided.
        """
        X = misc.validate_input_type(X, return_type="array").astype("float32")

        if self.verbose > 0:
            print(
                f"\nAdding {self.prop_missing} missing data per column "
                f"using strategy: {self.strategy}"
            )

        if np.all(np.isnan(np.array([self.missing_val])) == False):
            X[X == self.missing_val] = np.nan

        self.original_missing_mask_ = np.isnan(X)

        if self.strategy == "random":
            if self.mask_missing:
                # Get indexes where non-missing (Xobs) and missing (Xmiss).
                Xobs = np.where(~self.original_missing_mask_.ravel())[0]
                Xmiss = np.where(self.original_missing_mask_.ravel())[0]

                # Generate mask of 0's (non-missing) and 1's (missing).
                obs_mask = np.random.choice(
                    [0, 1],
                    size=Xobs.size,
                    p=((1 - self.prop_missing), self.prop_missing),
                ).astype(bool)

                # Make missing data mask.
                mask = np.zeros(X.size)
                mask[Xobs] = obs_mask
                mask[Xmiss] = 1

                # Reshape from raveled to 2D.
                # With strategy=="random", mask_ is equal to all_missing_.
                self.mask_ = np.reshape(mask, X.shape)

            else:
                # Generate mask of 0's (non-missing) and 1's (missing).
                self.mask_ = np.random.choice(
                    [0, 1],
                    size=X.shape,
                    p=((1 - self.prop_missing), self.prop_missing),
                ).astype(bool)

            # Make sure no entirely missing columns were simulated.
            self._validate_mask()

        elif self.strategy == "random_weighted":
            self.mask_ = self.random_weighted_missing_data(X, inv=False)

        elif self.strategy == "random_weighted_inv":
            self.mask_ = self.random_weighted_missing_data(X, inv=True)

        elif (
            self.strategy == "nonrandom"
            or self.strategy == "nonrandom_weighted"
        ):
            if self.genotype_data.tree is None:
                raise TypeError(
                    "SimGenotypeData.tree cannot be NoneType when "
                    "strategy='nonrandom' or 'nonrandom_weighted'"
                )

            mask = np.full_like(X, 0.0, dtype=bool)

            if self.strategy == "nonrandom_weighted":
                weighted = True
            else:
                weighted = False

            sample_map = dict()
            for i, sample in enumerate(self.genotype_data.samples):
                sample_map[sample] = i

            # if no tolerance provided, set to 1 snp position
            if self.tol is None:
                self.tol = 1.0 / mask.size

            # if no max_tries provided, set to # inds
            if self.max_tries is None:
                self.max_tries = mask.shape[0]

            filled = False
            while not filled:
                # Get list of samples from tree
                samples = self._sample_tree(
                    internal_only=False, skip_root=True, weighted=weighted
                )

                # Convert to row indices
                rows = [sample_map[i] for i in samples]

                # Randomly sample a column
                col_idx = np.random.randint(0, mask.shape[1])
                sampled_col = copy.copy(mask[:, col_idx])
                miss_mask = copy.copy(self.original_missing_mask_[:, col_idx])

                # Mask column
                sampled_col[rows] = True

                # If original was missing, set back to False.
                if self.mask_missing:
                    sampled_col[miss_mask] = False

                # check that column is not 100% missing now
                # if yes, sample again
                if np.sum(sampled_col) == sampled_col.size:
                    continue

                # if not, set values in mask matrix
                else:
                    mask[:, col_idx] = sampled_col

                    # if this addition pushes missing % > self.prop_missing,
                    # check previous prop_missing, remove masked samples from
                    # this column until closest to target prop_missing
                    current_prop = np.sum(mask) / mask.size
                    if abs(current_prop - self.prop_missing) <= self.tol:
                        filled = True
                        break
                    elif current_prop > self.prop_missing:
                        tries = 0
                        while (
                            abs(current_prop - self.prop_missing) > self.tol
                            and tries < self.max_tries
                        ):
                            r = np.random.randint(0, mask.shape[0])
                            c = np.random.randint(0, mask.shape[1])
                            mask[r, c] = False
                            tries += 1
                            current_prop = np.sum(mask) / mask.size

                        filled = True
                    else:
                        continue

            # With strategy=="nonrandom" or "nonrandom_weighted",
            # mask_ is equal to sim_missing_mask_ if mask_missing is True.
            # Otherwise it is equal to all_missing_.
            self.mask_ = mask

            self._validate_mask()

        else:
            raise ValueError(
                "Invalid SimGenotypeData.strategy value:", self.strategy
            )

        # Get all missing values.
        self.all_missing_mask_ = np.logical_or(
            self.mask_, self.original_missing_mask_
        )
        # Get values where original value was not missing and simulated.
        # data is missing.
        self.sim_missing_mask_ = np.logical_and(
            self.all_missing_mask_, self.original_missing_mask_ == False
        )

        self._validate_mask(mask=self.mask_missing)

        return self

    def transform(self, X):
        """Function to generate masked sites in a SimGenotypeData object

        Args:
            X (pandas.DataFrame, numpy.ndarray, or List[List[int]]): Data to transform. No missing data should be present in X. It should have already been imputed with one of the non-machine learning simple imputers.

        Returns:
            numpy.ndarray: Transformed data with missing data added.
        """
        X = misc.validate_input_type(X, return_type="array")

        # mask 012-encoded and one-hot encoded genotypes.
        return self._mask_snps(X)

    def accuracy(self, X_true, X_pred):
        """Calculate imputation accuracy of the simulated genotypes.

        Args:
            X_true (np.ndarray): True values.

            X_pred (np.ndarray): Imputed values.

        Returns:
            float: Accuracy score between X_true and X_pred.
        '"""
        masked_sites = np.sum(self.sim_missing_mask_)
        num_correct = np.sum(
            X_true[self.sim_missing_mask_] == X_pred[self.sim_missing_mask_]
        )
        return num_correct / masked_sites

    def auc_roc_pr_ap(self, X_true, X_pred):
        """Calcuate AUC-ROC, Precision-Recall, and Average Precision (AP).

        Args:
            X_true (np.ndarray): True values.

            X_pred (np.ndarray): Imputed values.

        Returns:
            List[float]: List of AUC-ROC scores in order of: 0,1,2.
            List[float]: List of precision scores in order of: 0,1,2.
            List[float]: List of recall scores in order of: 0,1,2.
            List[float]: List of average precision scores in order of 0,1,2.

        """
        y_true = X_true[self.sim_missing_mask_]
        y_pred = X_pred[self.sim_missing_mask_]

        # Binarize the output
        y_true_bin = label_binarize(y_true, classes=[0, 1, 2])
        y_pred_bin = label_binarize(y_pred, classes=[0, 1, 2])

        # Initialize lists to hold the scores for each class
        auc_roc_scores = []
        precision_scores = []
        recall_scores = []
        avg_precision_scores = []

        for i in range(y_true_bin.shape[1]):
            # AUC-ROC score
            auc_roc = roc_auc_score(
                y_true_bin[:, i], y_pred_bin[:, i], average="weighted"
            )
            auc_roc_scores.append(auc_roc)

            # Precision-recall score
            precision, recall, _, _ = precision_recall_fscore_support(
                y_true_bin[:, i], y_pred_bin[:, i], average="weighted"
            )
            precision_scores.append(precision)
            recall_scores.append(recall)

            # Average precision score
            avg_precision = average_precision_score(
                y_true_bin[:, i], y_pred_bin[:, i], average="weighted"
            )
            avg_precision_scores.append(avg_precision)

        return (
            auc_roc_scores,
            precision_scores,
            recall_scores,
            avg_precision_scores,
        )

    def random_weighted_missing_data(self, X, inv=False):
        """Choose values for which to simulate missing data by biasing towards the minority or majority alleles, depending on whether inv is True or False.

        Args:
            X (np.ndarray): True values.

            inv (bool, optional): If True, then biases towards choosing majority alleles. If False, then generates a stratified random sample (class proportions ~= full dataset) Defaults to False.

        Returns:
            np.ndarray: X with simulated missing values.

        """
        # Get unique classes and their counts
        classes, counts = np.unique(X, return_counts=True)
        # Compute class weights
        if inv:
            class_weights = 1 / counts
        else:
            class_weights = counts
        # Normalize class weights
        class_weights = class_weights / sum(class_weights)

        # Compute mask
        if self.mask_missing:
            # Get indexes where non-missing (Xobs) and missing (Xmiss)
            Xobs = np.where(~self.original_missing_mask_.ravel())[0]
            Xmiss = np.where(self.original_missing_mask_.ravel())[0]

            # Generate mask of 0's (non-missing) and 1's (missing)
            obs_mask = np.random.choice(
                classes, size=Xobs.size, p=class_weights
            )
            obs_mask = (obs_mask == classes[:, None]).argmax(axis=0)

            # Make missing data mask
            mask = np.zeros(X.size, dtype=bool)
            mask[Xobs] = obs_mask
            mask[Xmiss] = 1

            # Reshape from raveled to 2D
            mask = mask.reshape(X.shape)
        else:
            # Generate mask of 0's (non-missing) and 1's (missing)
            mask = np.random.choice(classes, size=X.size, p=class_weights)
            mask = (mask == classes[:, None]).argmax(axis=0).reshape(X.shape)

        # Assign mask to self before validation
        self.mask_ = mask

        self._validate_mask()

        return mask

    def _sample_tree(
        self,
        internal_only=False,
        tips_only=False,
        skip_root=True,
        weighted=False,
    ):
        """Function for randomly sampling clades from SimGenotypeData.tree.

        Args:
            internal_only (bool): Only sample from NON-TIPS. Defaults to False.

            tips_only (bool): Only sample from tips. Defaults to False.

            skip_root (bool): Exclude sampling of root node. Defaults to True.

            weighted (bool): Weight sampling by branch length. Defaults to False.

        Returns:
            List[str]: List of descendant tips from the sampled node.

        Raises:
            ValueError: ``tips_only`` and ``internal_only`` cannot both be True.
        """

        if tips_only and internal_only:
            raise ValueError("internal_only and tips_only cannot both be true")

        # to only sample internal nodes add  if not i.is_leaf()
        node_dict = dict()

        for node in self.genotype_data.tree.treenode.traverse("preorder"):
            ## node.idx is node indexes.
            ## node.dist is branch lengths.
            if skip_root:
                # If root node.
                if node.idx == self.genotype_data.tree.nnodes - 1:
                    continue

            if tips_only and internal_only:
                raise ValueError(
                    "tips_only and internal_only cannot both be True"
                )

            if tips_only:
                if not node.is_leaf():
                    continue
            elif internal_only:
                if node.is_leaf():
                    continue
            node_dict[node.idx] = node.dist
        if weighted:
            s = sum(list(node_dict.values()))
            # Node index / sum of node distances.
            p = [i / s for i in list(node_dict.values())]
            node_idx = np.random.choice(list(node_dict.keys()), size=1, p=p)[0]
        else:
            # Get missing choice from random clade.
            node_idx = np.random.choice(list(node_dict.keys()), size=1)[0]
        return self.genotype_data.tree.get_tip_labels(idx=node_idx)

    def _validate_mask(self, mask=False):
        """Make sure no entirely missing columns are simulated."""
        if mask is None:
            mask = self.mask_
        for i, column in enumerate(self.mask_.T):
            if mask:
                miss_mask = self.original_missing_mask_[:, i]
                col = column[~miss_mask]
                obs_idx = np.where(~miss_mask)
                idx = obs_idx[np.random.choice(np.arange(len(obs_idx)))]
            else:
                col = column
                idx = np.random.choice(np.arange(col.shape[0]))
            if np.sum(col) == col.size:
                self.mask_[idx, i] = False

    def _mask_snps(self, X):
        """Mask positions in SimGenotypeData.snps and SimGenotypeData.onehot"""
        if len(X.shape) == 3:
            # One-hot encoded.
            mask_val = [0.0, 0.0, 0.0, 0.0]
        elif len(X.shape) == 2:
            # 012-encoded.
            mask_val = -9
        else:
            raise ValueError(f"Invalid shape of input X: {X.shape}")

        Xt = X.copy()
        mask_boolean = self.mask_ != 0
        Xt[mask_boolean] = mask_val
        return Xt

    def write_mask(self, filename_prefix):
        """Write mask to file.

        Args:
            filename_prefix (str): Prefix for the filenames to write to.
        """
        np.save(filename_prefix + "_mask.npy", self.mask_)
        np.save(filename_prefix + "_original_missing_mask.npy", self.original_missing_mask_)

    def read_mask(self, filename_prefix):
        """Read mask from file.

        Args:
            filename_prefix (str): Prefix for the filenames to read from.

        Returns:
            tuple of np.ndarray: The read masks.
        """
        # Check if files exist
        if not os.path.isfile(filename_prefix + "_mask.npy"):
            raise FileNotFoundError(filename_prefix + "_mask.npy" + " does not exist.")
        if not os.path.isfile(filename_prefix + "_original_missing_mask.npy"):
            raise FileNotFoundError(filename_prefix + "_original_missing_mask.npy" + " does not exist.")

        # Load mask from file
        self.mask_ = np.load(filename_prefix + "_mask.npy")
        self.original_missing_mask_ = np.load(filename_prefix + "_original_missing_mask.npy")

        # Recalculate all_missing_mask_ from mask_ and original_missing_mask_
        self.all_missing_mask_ = np.logical_or(self.mask_, self.original_missing_mask_)

        return self.mask_, self.original_missing_mask_, self.all_missing_mask_


    @property
    def missing_count(self) -> int:
        """Count of masked genotypes in SimGenotypeData.mask

        Returns:
            int: Integer count of masked alleles.
        """
        return np.sum(self.mask_)

    @property
    def prop_missing_real(self) -> float:
        """Proportion of genotypes masked in SimGenotypeData.mask

        Returns:
            float: Total number of masked alleles divided by SNP matrix size.
        """
        return np.sum(self.mask_) / self.mask_.size



#!/usr/bin/env python
import sys
import os
import subprocess
import errno

import toytree
import toyplot.pdf
import pyvolve
import copy
import random
import re
import numpy as np
import pandas as pd

import matplotlib.pyplot as plt

def main():
	"""
	Using pyvolve and toytree to simulate data for PG-SUI

	Two ways to run:
	- Simulate SNPs along a 'pseudo-chromosome' from which SNPs are sampled
	- Simulate genes/ loci, sampling SNPs separately for each

	Pseudo-chromosome(s):
	Set num_loci = 1 (or number of desired chromosomes)
		loc_length = chromosome length (e.g., 50000)
		snps_per_locus = # total snps you want (e.g., 1000)
		make_gene_trees = False

	Separate loci (e.g., you want gene trees)
	Set num_loci = number of genes/ loci (e.g., 1000)
		loc_length = locus length (e.g., 500 or 1000)
		snps_per_locus = 1 (usually, but could be higher)
		make_gene_trees = True (if you want gene trees)

	If you just want a SNP matrix for testing PG-SUI, option 1
	is faster, and functionally not very different given the simulation
	model has no explicit mechanism of linkage (i.e. so site independence
	is true regardless). The second option will create
	a greater amount of rate heterogeneity, since each locus will be
	initialized with its own rate matrix (GTR or GTR+Gamma). In either case,
	setting write_gene_alignments = True will create sub-directories called
	'full_alignments/' containing the full sequences from with snps_per_locus
	number of SNPs will be sampled (without replacement) and concatenated to
	create the final outputs.

	NOTE: This script is not a part of the PG-SUI API, and is written
	for a single purpose, i.e., is not generalized beyond some options
	which can be manually set below. It is intended to provide transparency
	for the simulation process used in the PG-SUI manuscript, *not* as
	a flexible/ portable tool -- meaning a lot of things are hard-coded.

	"""
	seed=1234
	random.seed(seed)

	num_clades=4
	samples_per_clade=20
	num_loci=1000
	loc_length=250
	write_gene_alignments=False
	make_gene_trees=False
	make_guidetrees=True #set to true to run IQTREE on simulated SNP matrices
	keep_all=False #set to true to keep ALL iqtree outputs
	keep_report=True #set to true to keep .iqtree files
	get_siterates=True #set to true to infer site-specific rates in IQTREE
	snps_per_locus=1
	iqtree_bin="iqtree2"
	get_rates=True
	iq_procs=4

	###################

	if get_siterates and not make_guidetrees:
		print("ERROR: can't set get_siterates=True and make_guidetrees=False")
		print("Setting make_guidetrees=True and proceeding...")
		make_guidetrees=True

	clades=[]
	poplabels=[]
	indlabels=[]
	for i in range(num_clades):
		clades.append(("pop"+str(i)))
		for j in range(samples_per_clade):
			poplabels.append(("pop"+str(i)))
			indlabels.append(("pop"+str(i)+"_"+str(j)))
	outgroup = "pop"+str(num_clades-1)+"_"

	####### Varying clade vs. stem heights
	for clade_height in [0.001, 0.002, 0.003, 0.004, 0.005, 0.006, 0.007, 0.008, 0.009]:
		print("clade heights: ", clade_height)
		stem_height = np.around(0.01-clade_height, decimals=3)
		print("stem height: ", stem_height)

		#skeleton tree as newick
		skeleton_tree = toytree.rtree.unittree(ntips=num_clades,
									treeheight=stem_height,
									random_names=False,
									seed=random.randint(1, (sys.maxsize * 2 + 1))).write(tree_format=5)
		#grab newick trees for each clade
		pop_idx=0
		guidetree = skeleton_tree
		for clade in clades:
			clade_tree = toytree.rtree.unittree(ntips=samples_per_clade,
										treeheight=clade_height,
										random_names=False,
										seed=random.randint(1, (sys.maxsize * 2 + 1))).write(tree_format=5)
			clade_tree = clade_tree.replace(";","")
			for i in range(samples_per_clade):
				#indlabels.append((clade+"_"+str(j)))
				clade_tree = re.sub("r", (clade+"_"), clade_tree)
			guidetree = guidetree.replace(("r"+str(pop_idx)), clade_tree)
			pop_idx+=1

		base="c"+str(clade_height)+"_s"+str(stem_height)
		tobj=toytree.tree(guidetree, tree_format=0)


		#Set up directory structure for this set of tree params
		treeset_path = "sim_"+base
		if not os.path.exists(treeset_path):
			os.mkdir(treeset_path)

		#save guide trees
		basic_tree_plot(tobj, (treeset_path+"/"+base+"_guidetree.pdf"))
		tobj.write((treeset_path+"/"+base+"_guidetree.tre"), tree_format=5)


		######## With and without rate heterogeneity
		#NOTE: Run alignments through IQ-TREE to get optional
		#Rate matrix and site-specific mutation rates
		data = dict()
		for ind in indlabels:
			data[ind] = list()

		my_tree = pyvolve.read_tree(tree=guidetree)

		#for model in ["gtr","gtrgamma"]:
		for model in ["gtrgamma"]:
			model_outpath=treeset_path+"/"+base+"_"+model
			if not os.path.exists(model_outpath):
				os.mkdir(model_outpath)

			for locus in range(num_loci):
				print(locus)
				f = np.random.random(4)
				f /= f.sum()
				parameters = {
				"mu":
					{"AC": np.random.uniform(low=0.0, high=1.0),
					"AG": np.random.uniform(low=0.0, high=1.0),
					"AT": np.random.uniform(low=0.0, high=1.0),
					"CG": np.random.uniform(low=0.0, high=1.0),
					"CT": np.random.uniform(low=0.0, high=1.0),
					"GT": np.random.uniform(low=0.0, high=1.0)},
				"state_freqs":
					[f[0], f[1], f[2], f[3]]
				}
				if model == "gtr":
					#GTR model, without rate heterogeneity
					my_model = pyvolve.Model("nucleotide",
						parameters)
				else:
					my_model = pyvolve.Model("nucleotide",
						parameters,
						rate_factors = [
									np.random.uniform(low=0.1, high=0.7, size=1),
									np.random.uniform(low=0.5, high=1.2, size=1),
									np.random.uniform(low=1.0, high=1.8, size=1),
									np.random.uniform(low=1.5, high=5.0, size=1)
									],
						rate_probs = [0.4, 0.3, 0.2, 0.1] )
				if write_gene_alignments:
					fasta_outpath=model_outpath + "/full_alignments"
					if not os.path.exists(fasta_outpath):
						os.mkdir(fasta_outpath)
				else:
					fasta_outpath=model_outpath
				fastaout=fasta_outpath +"/"+ base+"_"+model+"_loc"+str(locus) + "_gene-alignment.fasta"
				#sample a gene alignment
				loc = sample_locus(my_tree, my_model, loc_length, snps_per_locus, fastaout)

				if loc:
					#sample SNP(s) from gene alignment
					sampled = sample_snp(read_fasta(fastaout), loc_length, snps_per_locus)
					if sampled is not None:
						data = add_locus(data,sampled)

					if not write_gene_alignments:
						os.remove(fastaout)
						if make_gene_trees:
							print("ERROR: Can't make gene trees when write_gene_alignments = False")
					elif make_gene_trees:
						run_iqtree(fastaout,
							iqtree_path=iqtree_bin,
							keep_all=keep_all,
							keep_report=keep_report,
							rates=get_siterates,
							procs=iq_procs)
						reroot_tree(tree=(fastaout+".treefile"),
							rooted=(fastaout+".rooted.tre"),
							outgroup_wildcard=outgroup)

			#write full SNP alignment & generate tree
			all_snp_out=model_outpath+"/"+base+"_"+model+"_base-snps-concat.phylip"
			write_phylip(data, all_snp_out)
			if make_guidetrees:
				run_iqtree(all_snp_out,
					iqtree_path=iqtree_bin,
					keep_all=keep_all,
					keep_report=keep_report,
					rates=get_siterates,
					procs=iq_procs)
				reroot_tree(tree=(all_snp_out+".treefile"),
					rooted=(all_snp_out+".rooted.tre"),
					outgroup_wildcard=outgroup)

			######## Varying introgression weight
			num_sampled_loci = len(data[indlabels[0]])
			#Modeled as contemporary exchange from pop2 -> pop1
			model_base=model_outpath+"/"+base+"_"+model

			for alpha in [0.1, 0.0, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]:
			#for alpha in [0.1, 0.5, 0.9]:
				alpha_base=model_base + "_i" + str(alpha)

				#TO-DO: write a logfile recording which loci are introgressed?
				source_pool = [indlabels[i] for i, pop in enumerate(poplabels) if pop == "pop2"]
				target_pool = [indlabels[i] for i, pop in enumerate(poplabels) if pop == "pop1"]
				if alpha == 0.0:
					introgressed_data = copy.copy(data)
				else:
					introgressed_data = hybridization(data,
											prob=alpha,
											source=source_pool,
											target=target_pool)

				introgessed_aln_out=alpha_base + "_base-snps-concat.phylip"
				write_phylip(introgressed_data, introgessed_aln_out)
				if make_guidetrees:
					run_iqtree(introgessed_aln_out,
							iqtree_path=iqtree_bin,
							keep_all=keep_all,
							keep_report=keep_report,
							rates=get_siterates,
							procs=iq_procs)
					reroot_tree(tree=(introgessed_aln_out+".treefile"),
						rooted=(introgessed_aln_out+".rooted.tre"),
						outgroup_wildcard=outgroup)

def reroot_tree(tree, rooted="out.rooted.tre", outgroup_wildcard="out"):
	t=toytree.tree(tree)
	try:
		rt=t.root(wildcard=outgroup_wildcard)
		rt.write(rooted, tree_format=5)
		return(rt)
	except Exception:
		t.write(rooted, tree_format=5)
		return(None)


def hybridization(dat, prob=0.1, source=None, target=None):
	new_dat=dict()
	if source is None:
		source = [key for key in dat.keys()]
	if target is None:
		target = [key for key in dat.keys()]

	for individual in dat.keys():
		new_dat[individual] = dat[individual]
		aln_len=len(dat[individual])
	all_indices=list(range(aln_len))
	num=int(aln_len*prob)

	for target_individual in target:
		snp_indices = np.random.choice(all_indices, size=num, replace=False)
		for index in snp_indices:
			source_ind=np.random.choice(source, size=1)[0]
			new_dat[target_individual][index] = new_dat[source_ind][index]
	return(new_dat)

def run_iqtree(aln,
	iqtree_path="iqtree",
	keep_all=False,
	keep_report=False,
	outgroup=None,
	rates=False,
	procs=4):
	#run
	cmd = [iqtree_path,
					"-s",
					str(aln),
					"-m",
					"GTR+I*G4",
					"-redo",
					"-T",
					str(procs)
					]
	if outgroup is not None:
		cmd.append("-o")
		cmd.append(str(outgroup))
	if rates:
		#use -wst (NOT -mlrate) if using iq-tree 1.6xx
		#cmd.append("-wsr")
		cmd.append("--mlrate")
	result = subprocess.run(cmd, capture_output=True, text=True)
	#print(result.stdout)
	#print(result.stderr)

	if not keep_all:
		#delete everything except treefile
		silentremove((aln + ".bionj"))
		silentremove((aln + ".ckp.gz"))
		silentremove((aln + ".log"))
		silentremove((aln + ".mldist"))
		silentremove((aln + ".uniqueseq.phy"))
		if not keep_report:
			silentremove((aln + ".iqtree"))
	return((aln + ".treefile"))

def add_locus(d, new):
	for sample in d.keys():
		for snp in new[sample]:
			d[sample].append(snp)
	return(d)

def write_fasta(seqs, fas):
	with open(fas, 'w') as fh:
		#Write seqs to FASTA first
		for a in seqs.keys():
			name = ">" + str(a) + "\n"
			seq = "".join(seqs[a]) + "\n"
			fh.write(name)
			fh.write(seq)
		fh.close()

def write_phylip(seqs, phy):
	#get header
	samps=0
	snps=None
	for key in seqs.keys():
		samps+=1
		if snps is None:
			snps = len(seqs[key])
		elif snps != len(seqs[key]):
			raise ValueError(("Error writing file"+phy+"- sequences not equal length\n"))
	with open(phy, 'w') as fh:
		header=str(samps)+"\t"+str(snps)+"\n"
		fh.write(header)
		#Write seqs to FASTA first
		for a in seqs.keys():
			line = str(a) + "\t" + "".join(seqs[a]) + "\n"
			fh.write(line)
		fh.close()

def read_phylip(phy):
	data = dict()
	header=True
	sample=None
	with open(phy, "r") as fin:
		for line in fin:
			line = line.strip()
			if not line:  # If blank line.
				continue
			else:
				if header==True:
					header=False
					continue
				else:
					stuff = line.split()
					data[stuff[0]] = stuff[1]
		fin.close()
		return(data)

def read_fasta(fasta):
	data = dict()
	header=False
	sample=None
	sequence=""
	with open(fasta, "r") as fin:
		for line in fin:
			line = line.strip()
			if not line:  # If blank line.
				continue
			if line[0] == ">":
				if sample:
					data[sample] = sequence
					sequence = ""
				sample=line[1:]
			else:
				sequence = sequence + line
		data[sample] = sequence
		fin.close()
		return(data)

def sample_snp(aln_dict, aln_len, snps_per_locus=1):
	snp_indices = []
	snp_aln = dict()
	if aln_len == 1:
		for sample in aln_dict.keys():
			snp_aln[sample] = aln_dict[sample][0]
		return(snp_aln)
	else:
		for sample in aln_dict.keys():
			snp_aln[sample] = []
	for i in range(aln_len):
		vars=[]
		for sample in aln_dict.keys():
			nuc=aln_dict[sample][i]
			if len(vars) == 0:
				vars.append(nuc)
			elif nuc not in vars:
				snp_indices.append(i)
				break
	if len(snp_indices) == 0:
		return(None)
	elif len(snp_indices) == 1:
		#sample them all
		for sample in aln_dict.keys():
			snp_aln[sample] = aln_dict[sample][snp_indices[0]]
	else:
		sampled_indices = np.random.choice(snp_indices, size=snps_per_locus, replace=False)
		for sample in aln_dict.keys():
			for i in sampled_indices:
				snp_aln[sample].append(aln_dict[sample][i])
	return(snp_aln)

def sample_locus(tree, model, gene_len=1000, num_snps=1, out="out.fasta"):
	try:
		my_partition = pyvolve.Partition(models = model, size=gene_len)
		my_evolver = pyvolve.Evolver(partitions = my_partition, tree = tree)
		my_evolver(seqfile = out,
			seqfmt = "fasta",
			ratefile=False,
			infofile=False)
		return(True)
	except Exception:
		return False

def silentremove(filename):
	try:
		os.remove(filename)
	except OSError as e: # this would be "except OSError, e:" before Python 2.6
		if e.errno != errno.ENOENT: # errno.ENOENT = no such file or directory
			raise # re-raise exception if a different error occurred

def get_tree_tips(tree):
	tips = re.split('[ ,\(\);]', tree)
	return([i for i in tips if i])

def basic_tree_plot(tree, out="out.pdf"):
	mystyle = {
		"edge_type": "p",
		"edge_style": {
			"stroke": toytree.colors[0],
			"stroke-width": 1,
		},
		"tip_labels_align": True,
		"tip_labels_style": {"font-size": "5px"},
		"node_labels": False,
		"tip_labels": True
	}

	canvas, axes, mark = tree.draw(
		width=400,
		height=600,
		**mystyle,
	)

	toyplot.pdf.render(canvas, out)

if __name__ == "__main__":
	main()



# Standard library imports
import errno
import gc
import math
import os
import pprint
import sys
from collections import Counter
from collections import defaultdict
from operator import itemgetter
from pathlib import Path
from statistics import mean, median
from contextlib import redirect_stdout
from typing import Optional, Union, List, Dict, Tuple, Any, Callable
from copy import deepcopy


# Third party imports
import numpy as np
import pandas as pd

from scipy import stats as st

# from memory_profiler import memory_usage

# Scikit-learn imports
from sklearn.experimental import enable_iterative_imputer
from sklearn import metrics

from sklearn_genetic.space import Continuous, Categorical, Integer

# Custom module imports
try:
    from .supervised.iterative_imputer_gridsearch import (
        IterativeImputerGridSearch,
    )
    from .supervised.iterative_imputer_fixedparams import (
        IterativeImputerFixedParams,
    )
    from .unsupervised.neural_network_imputers import VAE, UBP, SAE
    from ..utils.misc import isnotebook
    from ..utils.misc import timer
    from ..data_processing.transformers import (
        SimGenotypeDataTransformer,
    )
except (ModuleNotFoundError, ValueError, ImportError):
    from impute.supervised.iterative_imputer_gridsearch import (
        IterativeImputerGridSearch,
    )
    from impute.supervised.iterative_imputer_fixedparams import (
        IterativeImputerFixedParams,
    )
    from impute.unsupervised.neural_network_imputers import VAE, UBP, SAE
    from utils.misc import isnotebook
    from utils.misc import timer
    from data_processing.transformers import (
        SimGenotypeDataTransformer,
    )

is_notebook = isnotebook()

if is_notebook:
    from tqdm.notebook import tqdm as progressbar
else:
    from tqdm import tqdm as progressbar


class Impute:
    """Class to impute missing data from the provided classifier.

    The Impute class will either run a variational autoencoder or IterativeImputer with the provided estimator. The settings for the provided estimator should be provided as the ``kwargs`` argument as a dictionary object with the estimator's keyword arguments as the keys and the corresponding values. E.g., ``kwargs={"n_jobs", 4, "initial_strategy": "populations"}``\. ``clf_type`` just specifies either "classifier" or "regressor". "regressor" is primarily just for quick and dirty testing and is intended for internal use only.

    Once the Impute class is initialized, the imputation should be performed with ``fit_predict()``\.

    The imputed data can then be written to a file with ``write_imputed()``

    Args:
        clf (str or Callable estimator object): The estimator object to use. If using a variational autoencoder, the provided value should be "VAE". Otherwise, it should be a callable estimator object that is compatible with scikit-learn's IterativeImputer.

        clf_type (str): Specify whether to use a "classifier" or "regressor". The "regressor" option is just for quick and dirty testing, and "classifier" should almost always be used.

        kwargs (Dict[str, Any]): Settings to use with the estimator. The keys should be the estimator's keywords, and the values should be their corresponding settings.

    Raises:
        TypeError: Check whether the ``gridparams`` values are of the correct format if ``ga=True`` or ``ga=False``\.

    Examples:
        # Don't use parentheses after estimator object.
        >>>imputer = Impute(sklearn.ensemble.RandomForestClassifier,
        "classifier",
        {"n_jobs": 4, "initial_strategy": "populations", "max_iter": 25, "n_estimators": 100, "ga": True})
        >>>self.imputed, self.best_params = imputer.fit_predict(df)
        >>>imputer.write_imputed(self.imputed)
        >>>print(self.imputed)
        [[0, 0, 0, 0],
        [0, 0, 0, 0],
        [0, 1, 1, 0],
        [2, 1, 2, 2]]
    """

    def __init__(
        self, clf: Union[str, Callable], clf_type: str, kwargs: Dict[str, Any]
    ) -> None:
        self.clf = clf
        self.clf_type = clf_type
        self.original_num_cols = None

        if self.clf == VAE or self.clf == SAE or self.clf == UBP:
            self.algorithm = "nn"
            self.imp_method = "Unsupervised"
        else:
            self.algorithm = "ii"
            self.imp_method = "Supervised"

        self.imp_name = self.clf.__name__

        try:
            self.pops = kwargs["genotype_data"].populations
        except AttributeError:
            self.pops = None

        self.genotype_data = kwargs["genotype_data"]
        self.verbose = kwargs["verbose"]

        # Separate local variables into settings objects
        (
            self.imp_kwargs,
            self.clf_kwargs,
            self.ga_kwargs,
            self.cv,
            self.verbose,
            self.n_jobs,
            self.prefix,
            self.column_subset,
            self.disable_progressbar,
            self.chunk_size,
            self.do_validation,
            self.do_gridsearch,
            self.testing,
        ) = self._gather_impute_settings(kwargs)

        if self.algorithm == "ii":
            self.imp_kwargs["pops"] = self.pops

        if self.do_gridsearch:
            for v in kwargs["gridparams"].values():
                if (
                    isinstance(v, (Categorical, Integer, Continuous))
                    and kwargs["gridsearch_method"].lower()
                    != "genetic_algorithm"
                ):
                    raise TypeError(
                        "gridsearch_method argument must equal 'genetic_algorithm' if gridparams values are of type sklearn_genetic.space"
                    )

        self.logfilepath = os.path.join(
            f"{self.prefix}_output",
            "logs",
            self.imp_method,
            self.imp_name,
            f"imputer_progress_log.txt",
        )

        self.invalid_indexes = None

        # Remove logfile if exists
        try:
            os.remove(self.logfilepath)
        except OSError:
            pass

        Path(
            os.path.join(
                f"{self.prefix}_output",
                "plots",
                self.imp_method,
                self.imp_name,
            )
        ).mkdir(parents=True, exist_ok=True)

        Path(
            os.path.join(
                f"{self.prefix}_output", "logs", self.imp_method, self.imp_name
            )
        ).mkdir(parents=True, exist_ok=True)

        Path(
            os.path.join(
                f"{self.prefix}_output",
                "reports",
                self.imp_method,
                self.imp_name,
            )
        ).mkdir(parents=True, exist_ok=True)

        Path(
            os.path.join(
                f"{self.prefix}_output",
                "alignments",
                self.imp_method,
                self.imp_name,
            )
        ).mkdir(parents=True, exist_ok=True)

    @timer
    def fit_predict(
        self, X: pd.DataFrame
    ) -> Tuple[pd.DataFrame, Dict[str, Any]]:
        """Fit and predict imputations with IterativeImputer(estimator).

        Fits and predicts imputed 012-encoded genotypes using IterativeImputer with any of the supported estimator objects. If ``gridparams=None``\, then a grid search is not performed. If ``gridparams!=None``\, then a RandomizedSearchCV is performed on a subset of the data and a final imputation is done on the whole dataset using the best found parameters.

        Args:
            X (pandas.DataFrame): DataFrame with 012-encoded genotypes.

        Returns:
            GenotypeData: GenotypeData object with missing genotypes imputed.
            Dict[str, Any]: Best parameters found during grid search.
        """

        # Test if output file can be written to
        try:
            outfile = os.path.join(
                f"{self.prefix}_output",
                "alignments",
                self.imp_method,
                self.imp_name,
                "imputed_012.csv",
            )

            with open(outfile, "w") as fout:
                pass
        except IOError as e:
            print(f"Error: {e.errno}, {e.strerror}")
            if e.errno == errno.EACCES:
                sys.exit(f"Permission denied: Cannot write to {outfile}")
            elif e.errno == errno.EISDIR:
                sys.exit(f"Could not write to {outfile}; It is a directory")

        # mem_usage = memory_usage((self._impute_single, (X,)))
        # with open(f"profiling_results/memUsage_{self.prefix}.txt", "w") as fout:
        # fout.write(f"{max(mem_usage)}")
        # sys.exit()

        # Don't do a grid search
        if not self.do_gridsearch:
            imputed_df, df_scores, best_params = self._impute_single(X)

            if df_scores is not None:
                self._print_scores(df_scores)

        # Do a grid search and get the transformed data with the best parameters
        else:
            imputed_df, df_scores, best_params = self._impute_gridsearch(X)

            if self.verbose > 0:
                print("\nBest Parameters:")
                pprint.pprint(best_params)

        imp_data = self._imputed2genotypedata(imputed_df, self.genotype_data)

        print("\nDone!\n")
        return imp_data, best_params

    def _df2chunks(
        self, df: pd.DataFrame, chunk_size: Union[int, float]
    ) -> List[pd.DataFrame]:
        """Break up pandas.DataFrame into chunks and impute chunks.

        If set to 1.0 of type float, then returns only one chunk containing all the data.

        Args:
            df (pandas.DataFrame): DataFrame to split into chunks.

            chunk_size (int or float): If type is integer, then breaks DataFrame into ``chunk_size`` chunks. If type is float, breaks DataFrame up into ``chunk_size * len(df.columns)`` chunks.

        Returns:
            List[pandas.DataFrame]: List of pandas DataFrames of shape (n_samples, n_features_in_chunk).

        Raises:
            ValueError: ``chunk_size`` must be of type int or float.
        """
        if (
            "initial_strategy" in self.imp_kwargs
            and self.imp_kwargs["initial_strategy"] == "phylogeny"
            and chunk_size != 1.0
        ):
            print(
                "WARNING: Chunking is not supported with initial_strategy == "
                "'phylogeny'; Setting chunk_size to 1.0 and imputing entire "
                "dataset"
            )

            chunk_size = 1.0

        if (
            "initial_strategy" in self.imp_kwargs
            and self.imp_kwargs["initial_strategy"] == "mf"
            and chunk_size != 1.0
        ):
            print(
                "WARNING: Chunking is not supported with initial_strategy == "
                "'mf'; Setting chunk_size to 1.0 and imputing entire "
                "dataset"
            )

            chunk_size = 1.0

        if isinstance(chunk_size, (int, float)):
            chunks = list()
            df_cp = df.copy()

            if isinstance(chunk_size, float):
                if chunk_size > 1.0:
                    raise ValueError(
                        f"If chunk_size is of type float, must be "
                        f"between 0.0 and 1.0; Value supplied was {chunk_size}"
                    )

                elif chunk_size == 1.0:
                    # All data in one chunk
                    chunks.append(df_cp)
                    if self.verbose > 1:
                        print(
                            "Imputing all features at once since chunk_size is "
                            "set to 1.0"
                        )

                    return chunks

                tmp = chunk_size
                chunk_size = None
                chunk_size = math.ceil(len(df.columns) * tmp)

        else:
            raise ValueError(
                f"chunk_size must be of type float or integer, "
                f"but type {type(chunk_size)} was passed"
            )

        chunk_len_list = list()
        num_chunks = math.ceil(len(df.columns) / chunk_size)
        for i in range(num_chunks):
            chunks.append(df_cp.iloc[:, i * chunk_size : (i + 1) * chunk_size])
            chunk_len_list.append(len(chunks[i].columns))

        chunk_len = ",".join([str(x) for x in chunk_len_list])

        if self.verbose > 1:
            print(
                f"Data split into {num_chunks} chunks with {chunk_len} features"
            )

        return chunks

    def _imputed2genotypedata(self, imp012, genotype_data):
        """Create new instance of GenotypeData object from imputed DataFrame.

        The imputed, decoded DataFrame gets written to file and re-loaded to instantiate a new GenotypeData object.

        Args:
            imp012 (pandas.DataFrame): Imputed 012-encoded DataFrame.

            genotype_data (GenotypeData): Original GenotypeData object to load attributes from.

        Returns:
            GenotypeData: GenotypeData object with imputed data.
        """
        imputed_gd = deepcopy(genotype_data)

        # if self.clf == VAE:
        if len(imp012.shape) == 3:
            if imp012.shape[-1] == 4:
                imputed_gd.genotypes_onehot = imp012
            else:
                raise ValueError("Invalid shape for imputed output.")
        elif len(imp012.shape) == 2:
            if isinstance(imp012, pd.DataFrame):
                imp012 = imp012.to_numpy()
            imp012 = imp012.astype(int)
            if np.max(imp012) > 2:
                imputed_gd.genotypes_int = imp012
            else:
                imputed_gd.genotypes_012 = imp012
        else:
            raise ValueError(
                f"Invalid shape for imputed output: {imp012.shape}"
            )
        # else:
        #     imputed_gd.genotypes_012 = imp012

        return imputed_gd

    def _subset_data_for_gridsearch(
        self,
        df: pd.DataFrame,
        columns_to_subset: Union[int, float],
        original_num_cols: int,
    ) -> Tuple[pd.DataFrame, np.ndarray]:
        """Randomly subsets pandas.DataFrame.

        Subset pandas DataFrame with ``column_percent`` fraction of the data. Allows for faster validation.

        Args:
            df (pandas.DataFrame): DataFrame with 012-encoded genotypes.

            columns_to_subset (int or float): If float, proportion of DataFrame to randomly subset should be between 0 and 1. if integer, subsets ``columns_to_subset`` random columns.

            original_num_cols (int): Number of columns in original DataFrame.

        Returns:
            pandas.DataFrame: New DataFrame with random subset of features.
            numpy.ndarray: Sorted numpy array of column indices to keep.

        Raises:
            TypeError: column_subset must be of type float or int.
        """

        # Get a random numpy arrray of column names to select
        if isinstance(columns_to_subset, float):
            n = int(original_num_cols * columns_to_subset)
        elif isinstance(columns_to_subset, int):
            n = columns_to_subset
        else:
            raise TypeError(
                f"column_subset must be of type float or int, "
                f"but got {type(columns_to_subset)}"
            )

        col_arr = np.array(df.columns)

        if n > len(df.columns):
            if self.verbose > 0:
                print(
                    "Warning: Column_subset is greater than remaining columns following filtering. Using all columns"
                )

            df_sub = df.copy()
            cols = col_arr.copy()
        else:
            cols = np.random.choice(col_arr, n, replace=False)
            df_sub = df.loc[:, np.sort(cols)]
            # df_sub = df.sample(n=n, axis="columns", replace=False)

        df_sub.columns = df_sub.columns.astype(str)

        return df_sub, np.sort(cols)

    def _print_scores(self, df_scores: pd.DataFrame) -> None:
        """Print validation scores as pandas.DataFrame.

        Args:
            df (pandas.DataFrame): DataFrame with score statistics.
        """
        if self.verbose > 0:
            print("Validation scores:")
            print(df_scores)

    def _write_imputed_params_score(
        self, df_scores: pd.DataFrame, best_params: Dict[str, Any]
    ) -> None:
        """Save best_score and best_params to files on disk.

        Args:
            best_score (float): Best RMSE or accuracy score for the regressor or classifier, respectively.

            best_params (dict): Best parameters found in grid search.
        """

        best_score_outfile = os.path.join(
            f"{self.prefix}_output",
            "reports",
            self.imp_method,
            self.imp_name,
            "imputed_best_score.csv",
        )
        best_params_outfile = os.path.join(
            f"{self.prefix}_output",
            "reports",
            self.imp_method,
            self.imp_name,
            "imputed_best_params.csv",
        )

        if isinstance(df_scores, pd.DataFrame):
            df_scores.to_csv(
                best_score_outfile,
                header=True,
                index=False,
                float_format="%.2f",
            )

        else:
            with open(best_score_outfile, "w") as fout:
                fout.write(f"accuracy,{df_scores}\n")

        with open(best_params_outfile, "w") as fout:
            fout.write("parameter,best_value\n")
            for k, v in best_params.items():
                fout.write(f"{k},{v}\n")

    def _impute_single(
        self, df: pd.DataFrame
    ) -> Tuple[pd.DataFrame, pd.DataFrame, None]:
        """Run IterativeImputer without a grid search.

        Will do a different type of validation if ``do_validation == True``\.

        Args:
            df (pandas.DataFrame): DataFrame of 012-encoded genotypes.

        Returns:
            pandas.DataFrame: Imputed DataFrame of 012-encoded genotypes.
            pandas.DataFrame: DataFrame with validation scores.
            NoneType: Only used with _impute_gridsearch. Set to None here for compatibility.
        """
        if self.verbose > 0:
            print(
                f"\nDoing {self.clf.__name__} imputation without grid search..."
            )

        if self.algorithm == "nn":
            clf = None

        else:
            clf = self.clf(**self.clf_kwargs)

        if self.do_validation:
            if self.verbose > 0:
                print(f"Estimating {self.clf.__name__} validation scores...")

            if self.disable_progressbar:
                with open(self.logfilepath, "a") as fout:
                    # Redirect to progress logfile
                    with redirect_stdout(fout):
                        print(
                            f"Doing {self.clf.__name__} imputation "
                            f"without grid search...\n"
                        )

                        if self.verbose > 0:
                            print(
                                f"Estimating {self.clf.__name__} "
                                f"validation scores...\n"
                            )

            df_scores = self._imputer_validation(df, clf)

            if self.verbose > 0:
                print(f"\nDone with {self.clf.__name__} validation!\n")

            if self.disable_progressbar:
                if self.verbose > 0:
                    with open(self.logfilepath, "a") as fout:
                        # Redirect to progress logfile
                        with redirect_stdout(fout):
                            print(
                                f"\nDone with {self.clf.__name__} validation!\n"
                            )

        else:
            df_scores = None

        if self.algorithm == "nn":
            imputer = None

        else:
            imputer = self._define_iterative_imputer(
                clf,
                self.logfilepath,
                clf_kwargs=self.clf_kwargs,
                imp_kwargs=self.imp_kwargs,
            )

        if self.original_num_cols is None:
            self.original_num_cols = len(df.columns)

        # Remove non-biallelic loci
        # Only used if initial_strategy == 'phylogeny'
        if self.invalid_indexes is not None:
            df.drop(
                labels=self.invalid_indexes,
                axis=1,
                inplace=True,
            )

        if self.disable_progressbar:
            if self.verbose > 0:
                with open(self.logfilepath, "a") as fout:
                    # Redirect to progress logfile
                    with redirect_stdout(fout):
                        print(f"Doing {self.clf.__name__} imputation...\n")

        df_chunks = self._df2chunks(df, self.chunk_size)
        imputed_df = self._impute_df(df_chunks, imputer)
        imputed_df = imputed_df.astype(str)

        if self.disable_progressbar:
            if self.verbose > 0:
                with open(self.logfilepath, "a") as fout:
                    # Redirect to progress logfile
                    with redirect_stdout(fout):
                        print(f"\nDone with {self.clf.__name__} imputation!\n")

        lst2del = [df_chunks]
        del lst2del
        gc.collect()

        self._validate_imputed(imputed_df)

        if self.verbose > 0:
            print(f"\nDone with {self.clf.__name__} imputation!\n")

        return imputed_df, df_scores, None

    def _impute_gridsearch(
        self, df: pd.DataFrame
    ) -> Tuple[pd.DataFrame, pd.DataFrame, Dict[str, Any]]:
        """Do IterativeImputer with RandomizedSearchCV or GASearchCV.

        Args:
            df (pandas.DataFrame): DataFrame with 012-encoded genotypes.

        Returns:
            pandas.DataFrame: DataFrame with 012-encoded genotypes imputed using the best parameters found with the grid search.
            float: Absolute value of best score found during the grid search.
            dict: Best parameters found during the grid search.
        """
        original_num_cols = len(df.columns)
        df_subset, cols_to_keep = self._subset_data_for_gridsearch(
            df, self.column_subset, original_num_cols
        )

        print(f"Doing {self.clf.__name__} grid search...")

        if self.verbose > 0:
            print(f"Validation dataset size: {len(df_subset.columns)}\n")

        if self.disable_progressbar:
            with open(self.logfilepath, "a") as fout:
                # Redirect to progress logfile
                with redirect_stdout(fout):
                    print(f"Doing {self.clf.__name__} grid search...\n")

        if self.algorithm == "nn":
            self.imp_kwargs.pop("str_encodings")
            imputer = self.clf(
                **self.clf_kwargs,
                **self.imp_kwargs,
                ga_kwargs=self.ga_kwargs,
            )

            df_imp = pd.DataFrame(
                imputer.fit_transform(df_subset), columns=cols_to_keep
            )

            df_imp = df_imp.astype("float")
            df_imp = df_imp.astype("int64")

        else:
            clf = self.clf()
            df_subset = df_subset.astype("float32")
            df_subset.replace(-9.0, np.nan, inplace=True)

            imputer = self._define_iterative_imputer(
                clf,
                self.logfilepath,
                clf_kwargs=self.clf_kwargs,
                ga_kwargs=self.ga_kwargs,
                n_jobs=self.n_jobs,
                clf_type=self.clf_type,
                imp_kwargs=self.imp_kwargs,
            )

            if len(cols_to_keep) == original_num_cols:
                cols_to_keep = None

            Xt, params_list, score_list = imputer.fit_transform(
                df, cols_to_keep
            )

        if self.verbose > 0:
            print(f"\nDone with {self.clf.__name__} grid search!")

            if self.disable_progressbar:
                if self.verbose > 0:
                    with open(self.logfilepath, "a") as fout:
                        # Redirect to progress logfile
                        with redirect_stdout(fout):
                            print(
                                f"\nDone with {self.clf.__name__} grid search!"
                            )

        if self.algorithm == "ii":
            # Iterative Imputer.
            del imputer
            del Xt

            # Average or mode of best parameters
            # and write them to a file
            best_params = self._get_best_params(params_list)

            avg_score = mean(abs(x) for x in score_list if x != -9)
            median_score = median(abs(x) for x in score_list if x != -9)
            max_score = max(abs(x) for x in score_list if x != -9)
            min_score = min(abs(x) for x in score_list if x != -9)

            df_scores = pd.DataFrame(
                {
                    "Mean": avg_score,
                    "Median": median_score,
                    "Min": min_score,
                    "Max": max_score,
                },
                index=[0],
            )

            df_scores = df_scores.round(2)

            del avg_score
            del median_score
            del max_score
            del min_score
            gc.collect()
        else:
            # Using neural network.
            best_params = imputer.best_params_
            df_scores = imputer.best_score_
            df_scores = round(df_scores, 2) * 100
            best_imputer = None

        if self.clf_type == "classifier" and self.algorithm != "nn":
            df_scores = df_scores.apply(lambda x: x * 100)

        self._write_imputed_params_score(df_scores, best_params)

        # Change values to the ones in best_params
        self.clf_kwargs.update(best_params)

        if self.algorithm == "ii":
            if hasattr(self.clf(), "n_jobs"):
                self.clf_kwargs["n_jobs"] = self.n_jobs

            best_clf = self.clf(**self.clf_kwargs)

        gc.collect()

        if self.verbose > 0:
            print(
                f"\nDoing {self.clf.__name__} imputation "
                f"with best found parameters...\n"
            )

            if self.disable_progressbar:
                with open(self.logfilepath, "a") as fout:
                    # Redirect to progress logfile
                    with redirect_stdout(fout):
                        print(
                            f"\nDoing {self.clf.__name__} imputation "
                            f"with best found parameters...\n"
                        )

        if self.algorithm == "ii":
            best_imputer = self._define_iterative_imputer(
                best_clf,
                self.logfilepath,
                clf_kwargs=self.clf_kwargs,
                imp_kwargs=self.imp_kwargs,
            )

        final_cols = None
        if len(df.columns) < original_num_cols:
            final_cols = np.array(df.columns)

        if self.algorithm == "nn" and self.column_subset == 1.0:
            imputed_df = df_imp.copy()
            df_chunks = None
        else:
            df_chunks = self._df2chunks(df, self.chunk_size)
            imputed_df = self._impute_df(
                df_chunks, best_imputer, cols_to_keep=final_cols
            )

        lst2del = [df_chunks, df]
        del lst2del
        gc.collect()

        self._validate_imputed(imputed_df)

        if self.verbose > 0:
            print(f"Done with {self.clf.__name__} imputation!\n")

            if self.disable_progressbar:
                with open(self.logfilepath, "a") as fout:
                    # Redirect to progress logfile
                    with redirect_stdout(fout):
                        print(f"Done with {self.clf.__name__} imputation!\n")

        return imputed_df, df_scores, best_params

    def _imputer_validation(
        self, df: pd.DataFrame, clf: Optional[Callable]
    ) -> pd.DataFrame:
        """Validate imputation with a validation test set.

        Validation imputation by running it on a validation test set ``cv`` times. Actual missing values are imputed with sklearn.impute.SimpleImputer, and then missing values are randomly introduced to known genotypes. The dataset with no missing data is compared to the dataset with known missing data to obtain validation scores.

        Args:
            df (pandas.DataFrame): 012-encoded genotypes to impute.

            clf (sklearn classifier instance or None): sklearn classifier instance with which to run the imputation.

        Raises:
            ValueError: If none of the scores were able to be estimated and reps variable is empty.

        Returns:
            pandas.DataFrame: Validation scores in a pandas DataFrame object. Contains the scoring metric, mean, median, minimum, and maximum validation scores among all features, and the lower and upper 95% confidence interval among the replicates.
        """
        reps = defaultdict(list)
        for cnt, rep in enumerate(
            progressbar(
                range(self.cv),
                desc="Validation replicates: ",
                leave=True,
                disable=self.disable_progressbar,
            ),
            start=1,
        ):
            if self.disable_progressbar:
                perc = int((cnt / self.cv) * 100)
                if self.verbose > 0:
                    print(f"Validation replicate {cnt}/{self.cv} ({perc}%)")

                    with open(self.logfilepath, "a") as fout:
                        # Redirect to progress logfile
                        with redirect_stdout(fout):
                            print(
                                f"Validation replicate {cnt}/{self.cv} ({perc}%)"
                            )

            scores = self._impute_eval(df, clf)

            for k, score_list in scores.items():
                score_list_filtered = filter(lambda x: x != -9, score_list)

                if score_list_filtered:
                    reps[k].append(score_list_filtered)
                else:
                    continue

        if not reps:
            raise ValueError("None of the features could be validated!")

        ci_lower = dict()
        ci_upper = dict()
        for k, v in reps.items():
            reps_t = np.array(v).T.tolist()

            cis = list()
            if len(reps_t) > 1:
                for rep in reps_t:
                    rep = [abs(x) for x in rep]

                    cis.append(
                        st.t.interval(
                            alpha=0.95,
                            df=len(rep) - 1,
                            loc=np.mean(rep),
                            scale=st.sem(rep),
                        )
                    )

                ci_lower[k] = mean(x[0] for x in cis)
                ci_upper[k] = mean(x[1] for x in cis)
            else:
                print(
                    "Warning: There was no variance among replicates; "
                    "the 95% CI could not be calculated"
                )

                ci_lower[k] = np.nan
                ci_upper[k] = np.nan

        results_list = list()
        for k, score_list in scores.items():
            avg_score = mean(abs(x) for x in score_list if x != -9)
            median_score = median(abs(x) for x in score_list if x != -9)
            max_score = max(abs(x) for x in score_list if x != -9)
            min_score = min(abs(x) for x in score_list if x != -9)

            results_list.append(
                {
                    "Metric": k,
                    "Mean": avg_score,
                    "Median": median_score,
                    "Min": min_score,
                    "Max": max_score,
                    "Lower 95% CI": ci_lower[k],
                    "Upper 95% CI": ci_upper[k],
                }
            )

        df_scores = pd.DataFrame(results_list)

        if self.clf_type == "classifier":
            columns_list = [
                "Mean",
                "Median",
                "Min",
                "Max",
                "Lower 95% CI",
                "Upper 95% CI",
            ]

        df_scores = df_scores.round(2)

        outfile = os.path.join(
            f"{self.prefix}_output",
            "reports",
            self.imp_method,
            self.imp_name,
            "imputed_best_score.csv",
        )
        df_scores.to_csv(outfile, header=True, index=False)

        del results_list
        gc.collect()

        return df_scores

    def _impute_df(
        self,
        df_chunks: List[pd.DataFrame],
        imputer: Optional[
            Union[IterativeImputerFixedParams, IterativeImputerGridSearch]
        ] = None,
        cols_to_keep: Optional[np.ndarray] = None,
    ) -> pd.DataFrame:
        """Impute list of pandas.DataFrame objects using custom IterativeImputer class.

        The DataFrames are chunks of the whole input data, with each chunk correspoding to ``chunk_size`` features from ``_df2chunks()``\.

        Args:
            df_chunks (List[pandas.DataFrame]): List of Dataframes of shape(n_samples, n_features_in_chunk).

            imputer (imputer or classifier instance or None): Imputer or classifier instance to perform the imputation.

            cols_to_keep (numpy.ndarray or None): Final bi-allelic columns to keep. If some columns were non-biallelic, it will be a subset of columns.

        Returns:
            pandas.DataFrame: Single DataFrame object, with all the imputed chunks concatenated together.
        """
        imputed_chunks = list()
        for i, Xchunk in enumerate(df_chunks, start=1):
            if self.clf_type == "classifier":
                if self.algorithm == "nn":
                    if self.clf == VAE:
                        self.clf_kwargs["testing"] = self.testing
                    imputer = self.clf(
                        genotype_data=self.imp_kwargs["genotype_data"],
                        disable_progressbar=self.disable_progressbar,
                        prefix=self.prefix,
                        **self.clf_kwargs,
                    )
                    df_imp = pd.DataFrame(
                        imputer.fit_transform(Xchunk),
                    )
                    df_imp = df_imp.astype("float")
                    df_imp = df_imp.astype("Int8")

                else:
                    imp, _, __ = imputer.fit_transform(
                        Xchunk, valid_cols=cols_to_keep
                    )
                    df_imp = pd.DataFrame(imp)

                imputed_chunks.append(df_imp)

            else:
                # Regressor. Needs to be rounded to integer first.
                imp, _, __ = imputer.fit_transform(
                    Xchunk, valid_cols=cols_to_keep
                )
                df_imp = pd.DataFrame(imp)
                df_imp = df_imp.round(0).astype("Int8")

                imputed_chunks.append(df_imp)

        concat_df = pd.concat(imputed_chunks, axis=1)

        del imputed_chunks
        gc.collect()

        return concat_df

    def _validate_imputed(self, df: pd.DataFrame) -> None:
        """Asserts that there is no missing data left in the imputed DataFrame.

        Args:
            df (pandas.DataFrame): DataFrame with imputed 012-encoded genotypes.

        Raises:
            AssertionError: Error if missing values are still found in the dataset after imputation.
        """
        assert (
            not df.isnull().values.any()
        ), "Imputation failed...Missing values found in the imputed dataset"

    def _get_best_params(
        self, params_list: List[Dict[str, Any]]
    ) -> Dict[str, Any]:
        """[Gets the best parameters from the grid search. Determines the parameter types and either gets the mean or mode if the type is numeric or string/ boolean]

        Args:
            params_list (List[dict]): List of grid search parameter values.

        Returns:
            Dict[str, Any]: Dictionary with parameters as keys and their best values.
        """
        best_params = dict()
        keys = list(params_list[0].keys())
        first_key = keys[0]

        params_list = list(filter(lambda i: i[first_key] != -9, params_list))

        for k in keys:
            if all(
                isinstance(x[k], (int, float)) for x in params_list if x[k]
            ):
                if all(isinstance(y[k], int) for y in params_list):
                    best_params[k] = self._average_list_of_dicts(
                        params_list, k, is_int=True
                    )

                elif all(isinstance(z[k], float) for z in params_list):
                    best_params[k] = self._average_list_of_dicts(
                        params_list, k
                    )

            elif all(isinstance(x[k], (str, bool)) for x in params_list):
                best_params[k] = self._mode_list_of_dicts(params_list, k)

            else:
                best_params[k] = self._mode_list_of_dicts(params_list, k)

        return best_params

    def _mode_list_of_dicts(
        self, l: List[Dict[str, Union[str, bool]]], k: str
    ) -> str:
        """Get mode for key k in a list of dictionaries.

        Args:
            l (list(dict)): List of dictionaries.
            k (str): Key to find the mode across all dictionaries in l.

        Returns:
            str or bool: Most common value across list of dictionaries for one key.
        """
        k_count = Counter(map(itemgetter(k), l))
        return k_count.most_common()[0][0]

    def _average_list_of_dicts(
        self,
        l: List[Dict[str, Union[int, float]]],
        k: str,
        is_int: bool = False,
    ) -> Union[int, float]:
        """Get average of a given key in a list of dictionaries.

        Args:
            l (List[Dict[str, Union[int, float]]]): List of dictionaries.

            k (str): Key to find average across list of dictionaries.

            is_int (bool, optional): Whether or not the value for key k is an integer. If False, it is expected to be of type float. Defaults to False.

        Returns:
            int or float: average of given key across list of dictionaries.
        """
        if is_int:
            return int(sum(d[k] for d in l) / len(l))
        else:
            return sum(d[k] for d in l) / len(l)

    def _gather_impute_settings(
        self, kwargs: Dict[str, Any]
    ) -> Tuple[
        Optional[Dict[str, Any]],
        Optional[Dict[str, Any]],
        Optional[Dict[str, Any]],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[str],
        Optional[Union[int, float]],
        Optional[bool],
        Optional[Union[int, float]],
        Optional[bool],
        Optional[bool],
    ]:
        """Gather impute settings from kwargs object.

        Gather impute settings from the various imputation classes and IterativeImputer. Gathers them for use with the ``Impute`` class. Returns dictionary with keys as keyword arguments and the values as the settings. The imputation can then be run by specifying IterativeImputer(imp_kwargs).

        Args:
            kwargs (Dict[str, Any]): Dictionary with keys as the keyword arguments and their corresponding values.

        Returns:
            Dict[str, Any]: IterativeImputer keyword arguments.
            Dict[str, Any]: Classifier keyword arguments.
            Dict[str, Any]: Genetic algorithm keyword arguments.
            int: Number of cross-validation folds to use with non-grid search validation.
            int: Verbosity setting. 0 is silent, 2 is most verbose.
            int: Number of processors to use with grid search.
            str or None: Prefix for output files.
            int or float: Proportion of dataset (if float) or number of columns (if int) to use for grid search.
            bool: If True, disables the tqdm progress bar and just prints status updates to a file. If False, uses tqdm progress bar.
            int or float: Chunk sizes for doing full imputation following grid search. If int, then splits into chunks of ``chunk_size``\. If float, then splits into chunks of ``n_features * chunk_size``\.
            bool: Whether to do validation if ``gridparams is None``.
            bool: True if doing grid search, False otherwise.
        """
        n_jobs = kwargs.pop("n_jobs", 1)
        cv = kwargs.pop("cv", None)
        column_subset = kwargs.pop("column_subset", None)
        chunk_size = kwargs.pop("chunk_size", 1.0)
        do_validation = kwargs.pop("do_validation", False)
        verbose = kwargs.get("verbose", 0)
        disable_progressbar = kwargs.get("disable_progressbar", False)
        prefix = kwargs.get("prefix", "imputer")
        testing = kwargs.get("testing", False)
        do_gridsearch = False if kwargs["gridparams"] is None else True

        if prefix is None:
            prefix = "imputer"

        imp_kwargs = kwargs.copy()
        clf_kwargs = kwargs.copy()
        ga_kwargs = kwargs.copy()

        imp_keys = [
            "grid_iter",
            "tol",
            "verbose",
            "genotype_data",
            "str_encodings",
            "progress_update_percent",
            "sim_strategy",
            "sim_prop_missing",
            "gridparams",
            "gridsearch_method",
            "scoring_metric",
            "disable_progressbar",
            "prefix",
        ]

        if self.algorithm == "ii":
            imp_keys.extend(
                [
                    "n_nearest_features",
                    "max_iter",
                    "initial_strategy",
                    "imputation_order",
                    "skip_complete",
                    "random_state",
                    "sample_posterior",
                ]
            )

        ga_keys = [
            "population_size",
            "tournament_size",
            "elitism",
            "crossover_probability",
            "mutation_probability",
            "ga_algorithm",
        ]

        to_remove = ["self", "__class__"]

        for k, v in clf_kwargs.copy().items():
            if k in to_remove:
                clf_kwargs.pop(k)
            if k in imp_keys:
                clf_kwargs.pop(k)
            if k in ga_keys:
                clf_kwargs.pop(k)

        if "clf_random_state" in clf_kwargs:
            clf_kwargs["random_state"] = clf_kwargs.pop("clf_random_state")

        if "clf_tol" in clf_kwargs:
            clf_kwargs["tol"] = clf_kwargs.pop("clf_tol")

        for k, v in imp_kwargs.copy().items():
            if k not in imp_keys:
                imp_kwargs.pop(k)

        for k, v in ga_kwargs.copy().items():
            if k not in ga_keys:
                ga_kwargs.pop(k)

        if "ga_algorithm" in ga_kwargs:
            ga_kwargs["algorithm"] = ga_kwargs.pop("ga_algorithm")

        if self.clf_type == "regressor":
            ga_kwargs["criteria"] = "min"

        elif self.clf_type == "classifier":
            ga_kwargs["criteria"] = "max"

        return (
            imp_kwargs,
            clf_kwargs,
            ga_kwargs,
            cv,
            verbose,
            n_jobs,
            prefix,
            column_subset,
            disable_progressbar,
            chunk_size,
            do_validation,
            do_gridsearch,
            testing,
        )

    def _impute_eval(
        self, df: pd.DataFrame, clf: Optional[Callable]
    ) -> Dict[str, List[Union[float, int]]]:
        """Function to run IterativeImputer on a pandas.DataFrame.

        The dataframe columns are randomly subset and a fraction of the known, true values are converted to missing data to allow evalutation of the model with either accuracy or mean_squared_error scores.

        Args:
            df (pandas.DataFrame): Original DataFrame with 012-encoded genotypes.

            clf (sklearn Classifier or None): Classifier instance to use with IterativeImputer.

        Returns:
            Dict[List[float or int]]: Validation scores for the current imputation.
        """
        cols = np.random.choice(
            df.columns,
            int(len(df.columns) * self.column_subset),
            replace=False,
        )

        if self.verbose > 0:
            print(
                f"\nSimulating validation data with missing data proportion "
                f"{self.sim_prop_missing} and strategy {self.sim_strategy}"
            )

        df_known = df.copy()

        if self.algorithm == "nn":
            df_unknown = df_known.copy()

        else:
            df_unknown = pd.DataFrame(
                SimGenotypeDataTransformer(
                    self.genotype_data,
                    prop_missing=self.imp_kwargs["sim_prop_missing"],
                    strategy=self.imp_kwargs["sim_strategy"],
                ).fit_transform(df_known)
            )

        df_unknown_slice = df_unknown[cols]

        # Neural networks
        if self.algorithm == "nn":
            df_stg = df_unknown_slice.copy()

            for col in df_stg.columns:
                df_stg[col] = df_stg[col].replace({pd.NA: np.nan})
            # df_stg.fillna(-9, inplace=True)

            imputer = self.clf(
                prefix=self.prefix, **self.clf_kwargs, **self.imp_kwargs
            )

            df_imp = pd.DataFrame(
                imputer.fit_transform(df_stg.to_numpy()),
                columns=cols,
            )

            df_unknown_slice = pd.DataFrame(imputer.y_simulated_, columns=cols)
            df_known_slice = pd.DataFrame(imputer.y_original_, columns=cols)

            df_missing_mask = pd.DataFrame(
                imputer.sim_missing_mask_, columns=cols
            )

            df_imp = df_imp.astype("float")
            df_imp = df_imp.astype("int64")

        else:
            df_known_slice = df_known[cols]
            df_known_slice = df_known[cols]
            df_missing_mask = df_unknown_slice.isnull()

            df_unknown.replace(-9, np.nan, inplace=True)

            # Using IterativeImputer
            df_stg = df_unknown.copy()

            imputer = self._define_iterative_imputer(
                clf,
                self.logfilepath,
                clf_kwargs=self.clf_kwargs,
                imp_kwargs=self.imp_kwargs,
            )

            imp_arr = imputer.fit_transform(df_stg)

            # Get only subset of validation columns
            # get_loc returns the index of the value
            df_imp = pd.DataFrame(
                imp_arr[:, [df_unknown.columns.get_loc(i) for i in cols]],
                columns=cols,
            )

        # Get score of each column
        scores = defaultdict(list)
        for col in df_known_slice.columns:
            # Adapted from: https://medium.com/analytics-vidhya/using-scikit-learns-iterative-imputer-694c3cca34de

            mask = df_missing_mask[col]
            y_true = df_known[col]
            y_true = y_true[mask]

            y_pred = df_imp[col]
            y_pred = y_pred[mask]

            if self.clf_type == "classifier":
                if y_pred.empty:
                    scores["accuracy"].append(-9)
                    scores["precision"].append(-9)
                    scores["f1"].append(-9)
                    scores["recall"].append(-9)
                    scores["jaccard"].append(-9)
                    continue

                # Had to do this because get incompatible type error if using
                # initial_imputation="populations"
                if y_true.dtypes != "int64":
                    y_true = y_true.astype("int64")
                if y_pred.dtypes != "int64":
                    y_pred = y_pred.astype("int64")

                scores["accuracy"].append(
                    metrics.accuracy_score(y_true, y_pred)
                )

                scores["precision"].append(
                    metrics.precision_score(
                        y_true, y_pred, average="macro", zero_division=0
                    )
                )

                scores["f1"].append(
                    metrics.f1_score(
                        y_true, y_pred, average="macro", zero_division=0
                    )
                )

                scores["recall"].append(
                    metrics.recall_score(
                        y_true, y_pred, average="macro", zero_division=0
                    )
                )

                scores["jaccard"].append(
                    metrics.jaccard_score(
                        y_true, y_pred, average="macro", zero_division=0
                    )
                )

            else:
                scores["explained_var"].append(
                    metrics.explained_variance_score(y_true, y_pred)
                )

                scores["rmse"].append(
                    metrics.mean_squared_error(y_true, y_pred, squared=False)
                )

        lst2del = [
            df_stg,
            df_imp,
            df_known,
            df_known_slice,
            df_unknown,
        ]

        if self.algorithm == "nn":
            del lst2del
            del cols
        else:
            del lst2del
            del imp_arr
            del imputer
            del cols

        gc.collect()

        return scores

    def _define_iterative_imputer(
        self,
        clf: Callable,
        logfilepath: str,
        clf_kwargs: Optional[Dict[str, Any]] = None,
        imp_kwargs: Optional[str] = None,
        ga_kwargs: Optional[Dict[str, Any]] = None,
        n_jobs: Optional[int] = None,
        clf_type: Optional[str] = None,
    ) -> Union[IterativeImputerGridSearch, IterativeImputerFixedParams]:
        """Define an IterativeImputer instance.

        The instances are of custom, overloaded IterativeImputer classes.

        Args:
            clf (sklearn Classifier instance): Estimator to use with IterativeImputer.

            logfilepath (str): Path to progress log file.

            clf_kwargs (dict, optional): Keyword arguments for classifier. Defaults to None.

            imp_kwargs (Dict[str, Any], optional): Keyword arguments for imputation settings. Defaults to None.

            ga_kwargs (dict, optional): Keyword arguments for genetic algorithm grid search. Defaults to None.

            n_jobs (int, optional): Number of parallel jobs to use with the IterativeImputer grid search. Ignored if ``search_space=None``\. Defaults to None.

            clf_type (str, optional): Type of estimator. Valid options are "classifier" or "regressor". Ignored if ``search_space=None``\. Defaults to None.

        Returns:
            sklearn.impute.IterativeImputer: IterativeImputer instance.
        """
        if not self.do_gridsearch:
            imp = IterativeImputerFixedParams(
                logfilepath,
                clf_kwargs,
                estimator=clf,
                **imp_kwargs,
            )

        else:
            # Create iterative imputer
            imp = IterativeImputerGridSearch(
                logfilepath,
                clf_kwargs,
                ga_kwargs,
                estimator=clf,
                grid_n_jobs=n_jobs,
                clf_type=clf_type,
                **imp_kwargs,
            )

        return imp



## PG-SUI package by Bradley T. Martin and Tyler K. Chafin
## E-mail: evobio721@gmail.com
## Version 0.1, completed 13-Dec-2021

# Suppresses tensorflow GPU warnings.
import os
import warnings

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

from pgsui.utils.misc import get_processor_name


# Requires scikit-learn-intellex package
if get_processor_name().strip().startswith("Intel"):
    try:
        from sklearnex import patch_sklearn

        patch_sklearn()
        intelex = True
    except (ImportError, TypeError):
        warnings.warn(
            "Intel CPU detected but scikit-learn-intelex is not installed. We recommend installing it to speed up computation if your hardware supports it."
        )
        intelex = False
else:
    intelex = False

os.environ["intelex"] = str(intelex)

from pgsui.impute.estimators import (
    ImputeKNN,
    ImputeNLPCA,
    ImputeRandomForest,
    ImputeStandardAutoEncoder,
    ImputeUBP,
    ImputeVAE,
    ImputeXGBoost,
)

from pgsui.impute.simple_imputers import (
    ImputePhylo,
    ImputeMF,
    ImputeAlleleFreq,
    ImputeRefAllele,
)

from pgsui.data_processing.transformers import SimGenotypeDataTransformer

__all__ = [
    "ImputeUBP",
    "ImputeVAE",
    "ImputeXGBoost",
    "ImputeStandardAutoEncoder",
    "ImputeRandomForest",
    "ImputeNLPCA",
    "ImputeKNN",
    "SimGenotypeDataTransformer",
    "ImputePhylo",
    "ImputeMF",
    "ImputeAlleleFreq",
    "ImputeRefAllele",
]
